"""
Extension of libcloud's OpenStack Node Driver.
"""
import binascii
import copy
import random
import json
import os
import socket
import sys
import time
from datetime import datetime

from threepio import logger

import libcloud.compute.ssh

from libcloud.compute.types import Provider, NodeState, DeploymentError,\
    LibcloudError
from libcloud.compute.base import StorageVolume, VolumeSnapshot,\
    NODE_ONLINE_WAIT_TIMEOUT, SSH_CONNECT_TIMEOUT,\
    NodeAuthPassword, NodeDriver
from libcloud.compute.deployment import MultiStepDeployment, ScriptDeployment
from libcloud.compute.drivers.openstack import \
        OpenStack_1_1_NodeDriver,\
        OpenStack_1_1_Connection
from libcloud.utils.py3 import httplib
try:
    from lxml import etree as ET
except ImportError:
    from xml.etree import ElementTree as ET

from neutronclient.common.exceptions import NeutronClientException
from requests.exceptions import BaseHTTPError

from rfive.fabricSSH import FabricSSHClient

from rtwo.exceptions import NonZeroDeploymentException, ConnectionFailure
from rtwo.drivers.openstack_network import NetworkManager
from rtwo.drivers.openstack_user import UserManager
import functools

def swap_service_catalog(service_type=None, name=None):
    """
    Use this temporary decorator to take advantage of the full service catalog
    offered by OpenStack.
    TODO:Remove this when JMatts decorator is merged in!
    """
    def decorator(method):
        def service_catalog_switch(*args, **kwargs):
            """
            Switch service_catalog endpoints, then switch back!
            """
            driver = args[0]
            lc_conn = driver.connection
            if not lc_conn.service_catalog:
                lc_conn.get_service_catalog() # Make a request.
            if lc_conn._ex_force_base_url:
                old_endpoint = lc_conn._ex_force_base_url
            else:
                old_endpoint = lc_conn.get_endpoint()
            try:
                new_service = lc_conn.service_catalog.get_endpoint(
                    service_type=service_type, name=name,
                    region=lc_conn.service_region)
                if getattr(new_service, "url"):
                    lc_conn._ex_force_base_url = new_service.url
                else:
                    lc_conn._ex_force_base_url = new_service["url"]
                return method(*args, **kwargs)
            finally:
                lc_conn._ex_force_base_url = old_endpoint
        return service_catalog_switch
    return decorator

class OpenStack_Esh_Connection(OpenStack_1_1_Connection):
    #Ripped from OpenStackBaseConnection.__init__()
    def __init__(self, *args, **kwargs):
        timeout = kwargs.pop('timeout',None)
        #NOTE: If 'max attempts' is logic available in libcloud, Remove this and the 'request' logic, in favor of:
        # RETRY_FAILED_HTTP_REQUESTS: https://github.com/apache/libcloud/blob/trunk/libcloud/common/base.py#L76-L77
        # Be sure to set `retry_delay` and `backoff` in addition to `timeout`..
        self.max_attempts = kwargs.pop('max_attempts',2)
        if not timeout:
            timeout = 20 # Default 20 Second timeouts
        super(OpenStack_Esh_Connection, self).__init__(
            *args, timeout=timeout, **kwargs)

    def request(self, action, params=None,
                data='', headers=None, method='GET', max_attempts=None):
        if not max_attempts:
            max_attempts = self.max_attempts
        current_attempt = 0
        while current_attempt < max_attempts:
            try:
                current_attempt += 1
                response = super(OpenStack_1_1_Connection, self).request(
                        action=action,
                        params=params, data=data,
                        method=method, headers=headers)
                return response
            except (BaseHTTPError, httplib.HTTPException, socket.error,
                    socket.gaierror, httplib.BadStatusLine), e:
                _hostname = "%s:%s" % (self.host,self.port)
                logger.error("Request %s %s%s failed with error: %s - %s. Retry #%s/%s"
                        % (method, _hostname, action, e.__class__.__name__, e.args,
                           current_attempt, max_attempts))
                if current_attempt >= max_attempts:
                    logger.error("Final attempt failed! Request diagnostics:"
                            "URL:%s/%s, params=%s, data=%s,"
                            "method=%s, headers=%s"
                            % (_hostname, action, params, data, method, headers))
                    #This 3-tuple will re-raise the exception.
                    raise ConnectionFailure,\
                          "Final connection attempt exhausted:"\
                          " %s - %s" % (_hostname, e),\
                          sys.exc_info()[2]
                #DON'T FORGET TO WAIT BEFORE YOU RETRY! (4sec, 8sec)
                sleep_time = random.randint(0,2 ** current_attempt)
                time.sleep(sleep_time)
                logger.error("Waited %s seconds. Attempting again." % sleep_time)
            except Exception, e:
                raise

class OpenStack_Esh_NodeDriver(OpenStack_1_1_NodeDriver):
    """
    OpenStack node driver for esh.
    """
    connectionCls = OpenStack_Esh_Connection

    features = {
        "_to_volume": ["Convert native object to StorageVolume"],
        "_to_size": ["Add cpu info to extra, duplicate of vcpu"],
        "_to_image": ["Add state info to extra"],
        "_to_node": ["Build public_ips field",
                     "public_ips as extra",
                     "keypairs as extra",
                     "user/tenant as extra"],
        "create_node": ["Create node with ssh_key", "ssh_key",
                        "availability_zone"],
        "ex_create_node_with_network": ["Create node with floating IP"
                                        " and ssh_key", "ssh_key"],
        "ex_deploy_to_node": ["Deploy to existing node"],
        "ex_suspend_node": ["Suspends the node"],
        "ex_resume_node": ["Resume the node"],
        "ex_start_node": ["Starts the node"],
        "ex_stop_node": ["Stops the node"],
        "ex_vnc_console": ["Return a novnc token and url for a node."],
        "create_volume": ["Create volume"],
        "list_volumes": ["List all volumes"],
        "ex_list_floating_ip_pools": ["List all floating IP Pools"],
        "ex_delete_ports": ["Delete all ports associated with a node"],
        "ex_allocate_floating_ip": ["Allocate floating IP"],
        "ex_deallocate_floating_ip": ["Deallocate floating IP"],
        "ex_associate_floating_ip": ["Associate floating IP with node"],
        "ex_disassociate_floating_ip": ["Disassociate floating IP from node"],
        "ex_list_all_volumes": ["List all volumes for all tenants"
                                " for the user"],
        "ex_list_volume_attachments": ["List all attached volumes for node"],
        "ex_get_volume_attachment": ["Get details about an attached volume"],
        "ex_create_security_group": ["Add security group to tenant"],
        "ex_delete_security_group": ["Delete security group from tenant"],
        "ex_list_security_groups": ["List all security groups for tenant"],
        "ex_add_security_group": ["Add security group to tenant"],
        "ex_remove_security_group": ["Remove security group from tenant"],
        "ex_create_security_group_rule": ["Add rule to a group"],
        "ex_delete_security_group_rule": ["Remove rule from a group"],
        "ex_list_security_group_rules": ["List all rules for a group"],
        "ex_get_limits": ["Get Rate and Absolute API limits"],
        "ex_os_services": ["Manage services (os-services)"]
    }

    """
    Object builders -- Convert the native dict in to a Libcloud object
    """
    def _to_snapshot(self, api_ss):
        if 'snapshot' in api_ss:
            api_ss = api_ss['snapshot']

        extra = {'volume_id': api_ss.get('volumeId',api_ss.get('volume_id')),
                 'name': api_ss.get('displayName',api_ss.get('display_name')),
                 'created': api_ss.get('createdAt',api_ss.get('created_at')),
                 'description': api_ss.get('displayDescription', api_ss.get('display_description')),
                 'status': api_ss['status']} 
        snapshot = VolumeSnapshot(id=api_ss['id'], driver=self,
                                  size=api_ss['size'], extra=extra)
        return snapshot

    def _to_volumes(self, el, cinder=False):
        return [self._to_volume(volume, cinder=cinder)
                for volume in el['volumes']]

    def _cinder_volume_args(self, api_volume):
        api_volume['createdAt'] = api_volume.pop('created_at')
        api_volume['displayName']  = api_volume.pop('name', api_volume.pop('display_name','<No Name>'))
        api_volume['displayDescription'] = api_volume.pop('description', api_volume.pop('display_description',''))
        api_volume['availabilityZone'] = api_volume.pop('availability_zone')
        api_volume['snapshotId'] = api_volume.pop('snapshot_id')
        attachmentSet = api_volume['attachments']
        for attachment in attachmentSet:
            attachment['serverId'] = attachment.pop('server_id')
            attachment['volumeId'] = attachment.pop('volume_id')
        api_volume['attachments'] = attachmentSet
        return api_volume

    def _to_volume(self, api_volume, cinder=False):
        #Unwrap the object, if it wasn't unwrapped already.
        if not api_volume:
            return None
        if 'volume' in api_volume:
            api_volume = api_volume['volume']
        if cinder:
            api_volume = self._cinder_volume_args(api_volume)
        volume = super(OpenStack_Esh_NodeDriver, self)._to_volume(api_volume)

        created_time = datetime.strptime(api_volume['createdAt'], '%Y-%m-%dT%H:%M:%S.%f')
        volume.extra.update({
            'id': api_volume['id'],
            'object': api_volume,
            'displayName': api_volume['displayName'],
            'size': api_volume['size'],
            'status': api_volume['status'],
            'metadata': api_volume['metadata'],
            'availabilityZone': api_volume['availabilityZone'],
            'snapshotId': api_volume['snapshotId'],
            'createTime': created_time,
        })
        return volume

    def _to_size(self, api_size):
        """
        Extends Openstack_1_1_NodeDriver._to_size,
        adds support for cpu
        """
        size = super(OpenStack_Esh_NodeDriver, self)._to_size(api_size)
        size._api = api_size
        size.extra = {
            'cpu': api_size['vcpus'],
            'ephemeral': api_size.get('OS-FLV-EXT-DATA:ephemeral', 0),
            'public': api_size.get('os-flavor-access:is_public', True)
        }
        return size

    def _to_image(self, api_machine):
        """
        Extends Openstack_1_1_NodeDriver._to_image,
        adds support for architecture and state
        """
        #logger.debug(api_machine)
        image = super(OpenStack_Esh_NodeDriver, self)._to_image(api_machine)
        image.extra['state'] = api_machine['status'].lower()
        image.extra['image_size'] = api_machine['OS-EXT-IMG-SIZE:size']  # NOTE: This may change. keep an eye on this.. -Steve
        image.extra['api'] = api_machine
        return image

    def neutron_set_ips(self, node, floating_ips):
        """
        Using the network manager, find all IPs associated with this node
        """
        for f_ip in floating_ips:
            if f_ip.get('instance_id') == node.id:
                node.public_ips.append(f_ip['floating_ip_address'])
        return
    def neutron_list_networks(self, *args, **kwargs):
        """
        Although there is an 'os-networks' endpoint, valuable information like tenant_id are missing from the return-data, so we must call neutron directly..
        """
        network_manager = self.get_network_manager()
        networks = network_manager.list_networks(*args, **kwargs)
        return networks

    def neutron_get_tenant_network(self):
        network_manager = self.get_network_manager()
        tenant_networks = network_manager.tenant_networks()
        return tenant_networks[0] if tenant_networks else None

    def neutron_list_ports(self, *args, **kwargs):
        network_manager = self.get_network_manager()
        networks = network_manager.list_ports(*args, **kwargs)
        return networks

    def _to_nodes(self, el):
        return [self._to_node(api_node) for api_node in el['servers']]

    def _to_node(self, api_node, floating_ips=[]):
        """
        Extends OpenStack_1_1_NodeDriver._to_node
        adding support for public and private ips.
        """
        def _set_ips():
            """
            Set up ips in the api_node so _to_node may call its super.
            """
            try:
                public_ips, private_ips = [], []
                for (label, ip_addrs) in api_node['addresses'].items():
                    for ip in ip_addrs:
                        # If OS IP:type floating, assign to public network
                        # All other
                        if ip.get('OS-EXT-IPS:type') == 'floating':
                            public_ips.append(ip['addr'])
                        else:
                            private_ips.append(ip['addr'])
                [node.public_ips.append(ip) for ip in public_ips
                 if ip not in node.public_ips]
                [node.private_ips.append(ip) for ip in private_ips
                 if ip not in node.private_ips]
                #In this special case, it may be a matter of time before the ip
                #is available.. Atmosphere provides a hint with 'public-ip'
                #NOTE: Removed because it was causing problems in production
                #      cloud --Jan. 2015
                if node.private_ips and not node.public_ips \
                        and 'public-ip' in api_node['metadata']:
                    node.public_ips = [api_node['metadata']['public-ip']]
            except (IndexError, KeyError) as no_ip:
                logger.warn("No IP for node:%s" % api_node['id'])
        node = super(OpenStack_Esh_NodeDriver, self)._to_node(api_node)
        if floating_ips:
            self.neutron_set_ips(node, floating_ips)
        else:
            _set_ips()
        node.extra.update({
            'addresses': api_node.get('addresses'),
            'status': api_node.get('status').lower(),
            'fault': api_node.get('fault',{}),
            'task': api_node.get('OS-EXT-STS:task_state'),
            'power': api_node.get('OS-EXT-STS:power_state'),
            'object': api_node
        })
        return node

    def _copy_connection(self, **update_args):
        """
        Use this to copy the connectionObject
        """
        copied_args = {
            "user_id": self.connection.user_id,
            "key": self.key,
            "secret": self.secret,
            "secure": self.connection.secure,
            "host": self.connection.host,
            "port": self.connection.port,
            "timeout": self.connection.timeout,
            "ex_force_base_url": self._ex_force_base_url,
            "ex_force_auth_url": self._ex_force_auth_url,
            "ex_force_auth_version": self._ex_force_auth_version,
            "ex_force_auth_token": self._ex_force_auth_token,
            "ex_tenant_name": self._ex_tenant_name,
            "ex_force_service_type": self._ex_force_service_type,
            "ex_force_service_name": self._ex_force_service_name,
            "ex_force_service_region": self._ex_force_service_region,
        }
        copied_args.update(update_args)
        new_connection = self.__class__(**copied_args)
        return new_connection

    #def _make_keystone_connection(self):
    #    """
    #    Swap base url to make a request against keystone instead of nova
    #    """
    #    return self._copy_connection(
    #        ex_force_service_type='identity',
    #        ex_force_service_name='keystone')

    @swap_service_catalog(service_type="identity", name="keystone")
    def _keystone_list_users(self):
        user_resp = self.connection.request('/users').object
        all_users = user_resp['users']
        return all_users

    def _keystone_get_user(self, username):
        all_users = self._keystone_list_users()
        for user in all_users:
            if user['username'] == username or user['id'] == username:
                return user
        return None

    @swap_service_catalog(service_type="identity", name="keystone")
    def _keystone_list_tenants(self):
        tenant_resp = self.connection.request('/tenants').object
        all_tenants = tenant_resp['tenants']
        return all_tenants

    def _keystone_get_tenant(self, tenant_id_or_name):
        all_tenants = self._keystone_list_tenants()
        for tenant in all_tenants:
            if tenant['id'] == tenant_id_or_name or tenant['name'] == tenant_id_or_name:
                return tenant
        return None

    def _create_args_to_params(self, node, **kwargs):
        """
        NOTE: This is temporary. Latest version of libcloud has
        ex_availability_zone support
        """
        server_params = super(OpenStack_Esh_NodeDriver, self)\
            ._create_args_to_params(node, **kwargs)
        if 'ex_availability_zone' in kwargs:
            server_params['availability_zone'] = kwargs['ex_availability_zone']
        return server_params

    def _boot_volume_args_to_params(self, node, **kwargs):
        server_params = super(OpenStack_Esh_NodeDriver, self)\
            ._create_args_to_params(node, **kwargs)
        #Most of the work is taken care of at this stage...
        #But booting a volume requires some additional work.
        block_device_mapping = {
            "boot_index": kwargs.get("boot_index",0),
            "destination_type": kwargs.get("destination_type"),
            "delete_on_termination": kwargs.get("shutdown",False),
            "volume_size": kwargs.get("volume_size"),
            #NOTE: This line will likely change in icehouse. Replace it with:
            # "shutdown"  : kwargs.get("shutdown", False),
        }
        if kwargs.get('snapshot') and not node:
            block_device_mapping["source_type"] = "snapshot"
            block_device_mapping["uuid"] = kwargs['snapshot'].id
        elif kwargs.get('volume') and not node:
            block_device_mapping["source_type"] = "volume"
            block_device_mapping["uuid"] = kwargs['volume'].id
        elif kwargs.get('image') and not node:
            block_device_mapping["source_type"] = "image"
            block_device_mapping["uuid"] = kwargs['image'].id
        else:
            raise ValueError("To boot a volume, you must select a source."
                    " Available Sources: [image, volume, snapshot]")
            

        #NOTE: Wrapped in a list
        server_params["block_device_mapping_v2"] = [block_device_mapping]
        return server_params

    def create_node(self, **kwargs):
        """
        Helpful arguments to set:
        ex_keyname : Name of existing public key
        ex_availability_zone : Name of host to launch on
        ex_connection_kwargs : A dictionary of kwargs to be passed to the connection.
        """
        conn_kwargs = kwargs.pop('ex_connection_kwargs', {})
        server_params = self._create_args_to_params(None, **kwargs)

        resp = self.connection.request("/servers",
                                       method='POST',
                                       data={'server': server_params},
                                       **conn_kwargs)

        create_response = resp.object['server']
        server_resp = self.connection.request(
            '/servers/%s' % create_response['id'])
        server_object = server_resp.object['server']

        # adminPass is not always present
        # http://docs.openstack.org/essex/openstack-compute/admin/
        # content/configuring-compute-API.html#d6e1833
        server_object['adminPass'] = create_response.get('adminPass', None)

        node = self._to_node(server_object)

        return node

    def reboot_node(self, node, reboot_type='SOFT'):
        """
        Options for 'reboot_type': SOFT, HARD
        """
        return self._reboot_node(node, reboot_type=reboot_type)

    def ex_create_node_with_network(self, **kwargs):
        """
        Deprecated -- Old Workflow (Via JMATT!)
        """
        self._add_keypair(kwargs)
        kwargs.update({
            'ex_keyname': unicode(self.key),
        })
        logger.debug("kwargs = %s" % kwargs)
        #Instance launches at this point.
        node = super(OpenStack_Esh_NodeDriver, self).create_node(**kwargs)

        #NOTE: This line is needed to authenticate via SSH_Keypair instead!
        node.extra['password'] = None

        #NOTE: Using this to wait for the time it takes to launch
        # instance and have a valid IP port
        time.sleep(20)
        #TODO: It would be better to hook in an asnyc thread that
        # waits for valid IP port
        #TODO: This belongs in a eelery task.
        #server_id = node.id
        self.neutron_associate_ip(node, **kwargs)

        return node

    def ex_deploy_to_node(self, node, *args, **kwargs):
        """
        libcloud.compute.base.deploy_node
        """
        if not libcloud.compute.ssh.have_paramiko:
            raise RuntimeError('paramiko is not installed. You can install ' +
                               'it using pip: pip install paramiko')

        password = None

        if 'create_node' not in self.features:
            raise NotImplementedError(
                'deploy_node not implemented for this driver')
        elif 'generates_password' not in self.features["create_node"]:
            if 'password' not in self.features["create_node"] and \
               'ssh_key' not in self.features["create_node"]:
                raise NotImplementedError(
                    'deploy_node not implemented for this driver')

            if 'auth' not in kwargs:
                value = os.urandom(16)
                kwargs['auth'] = NodeAuthPassword(binascii.hexlify(value))

            if 'ssh_key' not in kwargs:
                password = kwargs['auth'].password

        max_tries = kwargs.get('max_tries', 3)

        if 'generates_password' in self.features['create_node']:
            password = node.extra.get('password')

        ssh_interface = kwargs.get('ssh_interface', 'public_ips')

        # Wait until node is up and running and has IP assigned
        try:
            node, ip_addresses = self.wait_until_running(
                nodes=[node],
                wait_period=3, timeout=NODE_ONLINE_WAIT_TIMEOUT,
                ssh_interface=ssh_interface)[0]
            if not ip_addresses:
                raise Exception('IP address was not found')
            logger.info("Ip Address found after calling wait_until_running: %s"
                        % ip_addresses)
        except Exception:
            e = sys.exc_info()[1]
            raise DeploymentError(node=node, original_exception=e, driver=self)

        if password:
            node.extra['password'] = password

        deploy_task = kwargs['deploy']
        ssh_username = kwargs.get('ssh_username', 'root')
        ssh_alternate_usernames = kwargs.get('ssh_alternate_usernames', [])
        ssh_port = kwargs.get('ssh_port', 22)
        ssh_timeout = kwargs.get('ssh_timeout', 10)
        ssh_key_file = kwargs.get('ssh_key', None)
        timeout = kwargs.get('timeout', SSH_CONNECT_TIMEOUT)

        deploy_error = None
        for username in ([ssh_username] + ssh_alternate_usernames):
            try:
                self._connect_and_run_deployment_script(
                    task=deploy_task, node=node,
                    ssh_hostname=ip_addresses[0], ssh_port=ssh_port,
                    ssh_username=username, ssh_password=password,
                    ssh_key_file=ssh_key_file, ssh_timeout=ssh_timeout,
                    timeout=timeout, max_tries=max_tries)
            except Exception as exc:
                # Try alternate username
                # TODO: Need to fix paramiko so we can catch a more specific
                # exception
                logger.exception("Could not connect to SSH on IP address %s" %
                                 ip_addresses[0])
                e = sys.exc_info()[1]
                deploy_error = e
            else:
                # Script sucesfully executed, don't try alternate username
                deploy_error = None
                break
        if deploy_error is not None:
            raise DeploymentError(node=node, original_exception=deploy_error,
                                  driver=self)
        if isinstance(deploy_task, ScriptDeployment):
            deploy_steps = [deploy_task]
        elif isinstance(deploy_task, MultiStepDeployment):
            deploy_steps = deploy_task.steps
        else:
            #No additional validation necessary for other types of deployment.
            return node
        #Additional validation for Script Deployments
        failed_steps = []
        for deploy_step in deploy_steps:
            if deploy_step.exit_status != 0:
                failed_steps.append(
                    "Script:%s returned a Non-Zero exit"
                    " status:%s. Stdout:%r Stderr:%r"
                    % (deploy_step.name, deploy_step.exit_status,
                       deploy_step.stdout, deploy_step.stderr))
        if failed_steps and not kwargs.get('non_zero_deploy'):
            raise DeploymentError(
                original_exception=NonZeroDeploymentException(
                    str(failed_steps)),
                node=node, driver=self)
        return node

    def _ssh_client_connect(self, ssh_client, wait_period=1.5, timeout=300):
        """
        Try to connect to the remote SSH server. If a connection times out or
        is refused it is retried up to timeout number of seconds.

        @keyword    ssh_client: A configured SSHClient instance
        @type       ssh_client: C{SSHClient}

        @keyword    wait_period: How many seconds to wait between each loop
                                 iteration (default is 1.5)
        @type       wait_period: C{int}

        @keyword    timeout: How many seconds to wait before timing out
                             (default is 600)
        @type       timeout: C{int}

        @return: C{SSHClient} on success
        """
        start = time.time()
        end = start + timeout

        while time.time() < end:
            try:
                ssh_client.connect(ignore_hosts=True)
            except (IOError, socket.gaierror, socket.error):
                # Retry if a connection is refused or timeout
                # occurred
                ssh_client.close()
                time.sleep(wait_period)
                continue
            else:
                return ssh_client

        raise LibcloudError(value='Could not connect to the remote SSH ' +
                            'server. Giving up.', driver=self)

    def _connect_and_run_deployment_script(self, task, node, ssh_hostname,
                                           ssh_port, ssh_username,
                                           ssh_password, ssh_key_file,
                                           ssh_timeout, timeout, max_tries):
        ssh_client = FabricSSHClient(hostname=ssh_hostname,
                                     port=ssh_port, username=ssh_username,
                                     password=ssh_password,
                                     key=ssh_key_file,
                                     timeout=ssh_timeout)

        # Connect to the SSH server running on the node
        logger.info(ssh_client.__dict__)
        ssh_client = self._ssh_client_connect(ssh_client=ssh_client,
                                              timeout=timeout)

        # Execute the deployment task
        self._run_deployment_script(task=task, node=node,
                                    ssh_client=ssh_client,
                                    max_tries=max_tries)

    def _run_deployment_script(self, task, node, ssh_client, max_tries=3):
        """
        Run the deployment script on the provided node. At this point it is
        assumed that SSH connection has already been established.

        @keyword    task: Deployment task to run on the node.
        @type       task: C{Deployment}

        @keyword    node: Node to operate one
        @type       node: C{Node}

        @keyword    ssh_client: A configured and connected SSHClient instance
        @type       ssh_client: C{SSHClientunlink

        @keyword    max_tries: How many times to retry if a deployment fails
                               before giving up (default is 3)
        @type       max_tries: C{int}

        @return: C{Node} Node instance on success.
        """
        tries = 0
        while tries < max_tries:
            try:
                node = task.run(node, ssh_client)
            except Exception:
                e = sys.exc_info()[1]
                tries += 1
                if tries >= max_tries:
                    e = sys.exc_info()[1]
                    raise LibcloudError(value='Failed after %d tries: %s'
                                        % (max_tries, str(e)), driver=self)
            else:
                return node
            finally:
                ssh_client.close()

    def ex_start_node(self, node):
        """
        Suspend a node.
        """
        resp = self._node_action(node, 'os-start')
        return resp.status == httplib.ACCEPTED

    def ex_stop_node(self, node):
        """
        Suspend a node.
        """
        resp = self._node_action(node, 'os-stop')
        return resp.status == httplib.ACCEPTED

    def ex_get_tenant_network(self, tenant_id):
        networks = self.ex_list_networks()
        tenant_network = [net for net in networks if tenant_id == net.tenant_id]
        return tenant_network

    def ex_suspend_node(self, node):
        """
        Suspend a node.
        """
        resp = self._node_action(node, 'suspend')
        return resp.status == httplib.ACCEPTED

    def ex_reset_network(self, node):
        """
        Resume a node.
        """
        resp = self._node_action(node, 'resetNetwork')
        return resp.status == httplib.ACCEPTED

    def ex_resume_node(self, node):
        """
        Resume a node.
        """
        resp = self._node_action(node, 'resume')
        return resp.status == httplib.ACCEPTED

    def ex_vnc_console(self, node, vnc_type='novnc'):
        """
        Return a novnc token and url for a node.
        Optional vnc_types:
        * novnc (Default - For Web Clients)
        * xvpvnc (For Java Clients)
        """
        resp = self._node_action(node,
                                 'os-getVNCConsole',
                                 type=vnc_type)
        return json.loads(resp.body)['console']['url']

    #quotas
    def _establish_connection(self):
        """
        This function will contact keystone for authorization
        and make an empty request to the server.

        Doing this will provide self.connection.auth_token.
        """
        try:
            self.connection.request('')
        except:
            #Will fail,but we MUST make a request to authenticate
            pass

    def _get_username(self):
        #if not self.connection.auth_token:
        #    self._establish_connection()
        return self.key

    def _get_user_id(self):
        if not self.connection.auth_token:
            self._establish_connection()
        return self.connection.auth_token.get('id')

    def _get_tenant_id(self):
        """
        After a successful auth, the 'action' will be:
        /v2/<tenant_id>
        We are parsing the action to retrieve the tenant_id.
        """
        if not self.connection.auth_token:
            self._establish_connection()
        action_str = self.connection.action
        if not action_str or len(action_str.split('/')) < 2:
            return None
        return self.connection.action.split("/")[2]

    #Volume Snapshots
    def ex_list_snapshots(self):
        """
        List the current users snapshots
        """
        server_resp = self.connection.request(
                '/os-snapshots')
        return [self._to_snapshot(snapshot)
                for snapshot in server_resp.object['snapshots']]

    @swap_service_catalog(service_type="volume", name="cinder")
    def ex_create_snapshot(self, display_name, display_description,
                           volume_id, force=True, tenant_id=None):
        """
        """
        body = {
            'snapshot': {
                "display_name": display_name,
                "display_description": display_description,
                "volume_id": volume_id,
                "force": force
            }
        }
        if tenant_id:
            body['tenant_id'] = tenant_id
        server_resp = self.connection.request(
                '/snapshots',
                data=body,
                method='POST')
        return self._to_snapshot(server_resp.object)

    @swap_service_catalog(service_type="volume", name="cinder")
    def ex_list_all_snapshots(self):
        """
        Admins only
        """
        server_resp = self.connection.request(
                '/snapshots/detail?all_tenants=1')
        return [self._to_snapshot(snapshot)
                for snapshot in server_resp.object['snapshots']]

    def ex_get_snapshot(self, snapshot_id):
        server_resp = self.connection.request(
                '/os-snapshots/%s' % snapshot_id)
        return self._to_snapshot(server_resp.object)

    #Volumes
    def ex_boot_volume(self, **kwargs):
        """
        """
        #Strict Validation required for some values..
        #1. if booting by something that is NOT a volume (Snapshot/Image)
        #   the size of the new volume must be created.
        logger.debug("driver.ex_boot_volume kwargs: %s" % kwargs)

        if not kwargs.get('volume') and not kwargs.get('volume_size'):
            raise ValueError(
                    "Conflict: Must define an explicit volume_size "
                    "when the source is not a volume")

        #Until we have a reason to support the 'local' and 'None' destination..
        if not kwargs.has_key("destination_type"):
            kwargs['destination_type'] = 'volume'

        server_params = self._boot_volume_args_to_params(None, **kwargs)
        server_resp = self.connection.request('/os-volumes_boot',
                                              method='POST',
                                              data={'server': server_params})
        return (server_resp.status == 200, server_resp.object)

    @swap_service_catalog(service_type="volume", name="cinder")
    def create_volume(self, size, name,
                      description=None, metadata=None,
                      location=None, snapshot=None, image=None, **connection_kwargs):
        """
        Create a new volume

        @keyword size: The size of the new volume
        @type    size: C{int}

        @keyword name: The name of the new volume
        @type    name: C{str}

        @keyword description: A description for the new volume (Optional)
        @type    description: C{str}

        @param   metadata: Key/Value metadata to associate with a node
        @type    metadata: C{dict}

        @keyword location: The location to place the new volume (Optional)
        @type    location: C{str}

        @keyword snapshot: Create a new volume from existing snapshot (Optional)
        @type    snapshot: C{VolumeSnapshot}

        @keyword image: Create a new volume from existing image (Optional)
        @type    image: C{Image}
        """
        body = {'volume': {
            'size': size,
            'display_name': name,
            }
        }
        if description:
            body['volume']['display_description'] = description
        if location:
            body['volume']['availability_zone'] = location
        if snapshot:
            body['volume']['snapshot_id'] = snapshot.id
        if image:
            body['volume']['imageRef'] = image.id
        server_resp = self.connection.request('/volumes',  # v3 doesnt use os-volumes anymore
                                              method='POST',
                                              data=body, **connection_kwargs)
        volume_obj =  self._to_volume(server_resp.object, cinder=True)
        return (server_resp.success(), volume_obj)

    def list_volumes(self):
        return self._to_volumes(self.connection.request("/os-volumes").object)

    def ex_volume_attached_to_instance(self, volume, instance_id):
        if not volume:
            return False
        attach_data = volume.extra.get('attachments', [])
        for attachment in attach_data:
            attached_instance_id = attachment.get('serverId')
            if not attached_instance_id:
                continue
            if attached_instance_id == instance_id:
                return True
        return False

    def ex_list_all_sizes(self):
        """
        List all instances from all tenants of a user
        """
        server_resp = self.connection.request(
            '/flavors/detail?all_tenants=1',
            method='GET')
        return self._to_sizes(server_resp.object)

    def ex_list_all_instances(self):
        """
        List all instances from all tenants of a user
        """
        #NOTE: THIS IS A HACK -- The 'admin' user should be able to see "All the things" -- HOWEVER
        # In the current implementation of liberty on jetstream, a call to 'list_all_tenants'
        # Made by a user with a single tenant will produce *IDENTICAL* results to that same call made by admin.
        # THIS IS CONSIDERED HARMFUL! So we have blocked all users except the admin accounts from making this call.
        if self.key not in ['atmoadmin','admin']:
            all_tenants = ""
        else:
            all_tenants = "?all_tenants=1"
        server_resp = self.connection.request(
            '/servers/detail%s' % all_tenants,
            method='GET')
        return self._to_nodes(server_resp.object)

    @swap_service_catalog(service_type="network", name="neutron")
    def _neutron_delete_quota(self, tenant_id):
        if not tenant_id:
            raise Exception("Tenant ID required to delete neutron quota")
        server_resp = self.connection.request(
            '/v2.0/quotas/%s.json' % tenant_id,
            method='DELETE')
        return server_resp.status == 204

    @swap_service_catalog(service_type="network", name="neutron")
    def _neutron_update_quota(self, tenant_id, new_values):
        if not tenant_id:
            raise Exception("Tenant ID required to update neutron quota")
        # Check for 'quota' wrapper
        if 'quota' not in new_values:
            new_values = {'quota': new_values}
        resp = self.connection.request(
            '/v2.0/quotas/%s.json' % tenant_id,
            data=new_values,
            method='PUT')
        quota = resp.object['quota']
        return quota

    @swap_service_catalog(service_type="network", name="neutron")
    def _neutron_show_quota(self, tenant_id=None):
        # If no tenant ID - use your own tenant ID
        tenant_resp = self.connection.request('/v2.0/quotas/tenant.json')
        tenant_obj = tenant_resp.object
        tenant_id = tenant_obj.get('tenant', {}).get('tenant_id', None)
        if not tenant_id:
            raise Exception("Error calling /v2.0/quotas/tenant.json - %s" % tenant_obj)
        # call to show quota:
        resp = self.connection.request('/v2.0/quotas/%s.json' % tenant_id)
        quota = resp.object['quota']
        return quota

    @swap_service_catalog(service_type="volume", name="cinder")
    def _cinder_delete_quota(self, username):
        lc_conn = self.connection
        server_resp = lc_conn.request(
            '/os-quota-sets/%s' % username,
            method='DELETE')
        return server_resp.status == 204

    @swap_service_catalog(service_type="volume", name="cinder")
    def _cinder_update_quota(self, username, new_values):
        lc_conn = self.connection
        # Check for 'quota_set' wrapper
        if 'quota_set' not in new_values:
            new_values = {'quota_set': new_values}
        server_resp = lc_conn.request(
            '/os-quota-sets/%s' % username,
            data=new_values,
            method='PUT')
        return server_resp.object.get('quota_set',{})


    @swap_service_catalog(service_type="volume", name="cinder")
    def _cinder_show_quota(self, username):
        lc_conn = self.connection
        server_resp = lc_conn.request(
            '/os-quota-sets/%s' % username,
            method='GET')
        return server_resp.object.get('quota_set',{})


    @swap_service_catalog(service_type="volume", name="cinder")
    def ex_list_all_volumes(self):
        lc_conn = self.connection
        server_resp = lc_conn.request(
            '/volumes/detail?all_tenants=1',
            method='GET')
        return self._to_volumes(server_resp.object, cinder=True)


    @swap_service_catalog(service_type="volume", name="cinder")
    def ex_update_volume(self, volume, **volume_updates):
        """
        Updates the editable attributes of a volume,
        Including: display_name, display_description
        For metadata, see 'ex_update_volume_metadata'
        """
        #ARGS formatting
        if 'name' in volume_updates:
            volume_updates['display_name'] = \
                    volume_updates.pop('name')
        elif 'displayName' in volume_updates:
            volume_updates['display_name'] = \
                    volume_updates.pop('displayName')
        if 'description' in volume_updates:
            volume_updates['display_description'] = \
                    volume_updates.pop('description')
        elif 'displayDescription' in volume_updates:
            volume_updates['display_description'] = \
                    volume_updates.pop('displayDescription')

        server_resp = self.connection.request(
                '/volumes/%s' % volume.id,
                method='PUT', data={'volume':volume_updates},
                )
        return self._to_volume(server_resp.object['volume'], cinder=True)

    @swap_service_catalog(service_type="volume", name="cinder")
    def ex_update_volume_metadata(self, volume, metadata):
        """
        Volume Metadata update
        metadata == dict of key/value metadata to be associated
        """
        data_dict = {'metadata': metadata}
        server_resp = self.connection.request('/volumes/%s/metadata' % volume.id,
                                              method='PUT',
                                              data=data_dict)
        try:
            return (server_resp.status == 200, server_resp.object['metadata'])
        except Exception, e:
            logger.exception("Exception occured updating volume")
            return (False, None)

    @swap_service_catalog(service_type="volume", name="cinder")
    def ex_delete_volume_metadata_key(self, volume, metadata_key):
        """
        """
        server_resp = self.connection.request(
                '/volumes/%s/metadata/%s' % (volume.id, metadata_key),
                method='DELETE')
        try:
            return server_resp.status == 200
        except Exception, e:
            logger.exception("Exception occured Removing Volume Metadata."
                             " Offending Key=%s" % metadata_key)
            return False


    def ex_list_volume_attachments(self, node):
        """
        List all attached/attaching volumes for a specific node
        """
        server_id = node.id
        server_resp = self.connection.request(
            '/servers/%s/os-volume_attachments' % server_id,
            method='GET')
        return server_resp.object.get('volumeAttachments', {})

    def ex_get_volume_attachment(self, volume):
        """
        Get details for specific volume attachment on a node
        """
        server_id = volume.attachment_set[0].get('serverId')
        attachment_id = volume.attachment_set[0].get('id')
        server_resp = self.connection.request(
            '/servers/%s/os-volume_attachments/%s' %
            (server_id, attachment_id),
            method='GET')
        return server_resp.object

    def ex_lookup_hypervisor_id_by_name(self, hypervisor_name):
        matches = [hv['id'] for hv in self.ex_list_hypervisor_nodes()
                   if hv['hypervisor_hostname'] == hypervisor_name]
        if not matches:
            raise ValueError("Hypervisor name %s has no corresponding ID.")
        return matches[0]

    def ex_list_instances_on_node(self, node_id):
        if type(node_id) == str:
            node_id = self.ex_lookup_hypervisor_id_by_name(node_id)
        return self.connection.request(
            "/os-hypervisors/%s/servers" % node_id).object['hypervisors']

    def ex_list_hypervisor_nodes(self):
        return self.connection.request(
            "/os-hypervisors").object['hypervisors']

    def ex_detail_hypervisor_nodes(self):
        return self.connection.request(
            "/os-hypervisors/detail").object['hypervisors']

    def ex_detail_hypervisor_node(self, node_id):
        if type(node_id) == str:
            node_id = self.ex_lookup_hypervisor_id_by_name(node_id)
        return self.connection.request(
            "/os-hypervisors/%s" % node_id).object['hypervisor']

    def ex_hypervisor_statistics(self):
        return self.connection.request(
            "/os-hypervisors/statistics").object['hypervisor_statistics']

    #Floating IPs
    def ex_list_floating_ips(self, region=None, **kwargs):
        """
        List all floating IPs in the tenants pool
        """
        def _to_ips(ip_list):
            return [floating_ip for floating_ip in ip_list['floating_ips']]
        try:
            return _to_ips(self.connection.request("/os-floating-ips").object)
        except:
            logger.warn("Unable to list floating ips from nova.")
            return []

    def ex_allocate_floating_ip(self, pool_name, **kwargs):
        """
        Allocate a new floating IP address to the tenants pool
        """
        try:
            floating_ip_obj = self.connection.request(
                '/os-floating-ips', method='POST',
                data={'pool': pool_name}).object
            return floating_ip_obj['floating_ip']
        except Exception, e:
            raise

    def ex_deallocate_floating_ip(self, floating_ip, **kwargs):
        """
        Deallocate an existing floating_ip from tenants pool
        """
        try:
            server_resp = self.connection.request(
                '/os-floating-ips/%s' % floating_ip,
                method='DELETE')
            return server_resp.object
        except Exception, e:
            raise

    def ex_clean_floating_ip(self, **kwargs):
        """
        Check for floating IPs without an instance ID
        and remove them from the driver
        """
        count = 0
        for f_ip in self.ex_list_floating_ips():
            if not f_ip.get('instance_id'):
                self.ex_deallocate_floating_ip(f_ip['id'])
                count += 1
        return count

    def ex_bulk_delete_floating_ips(self, ip_range):
        """
        Deallocate an existing floating_ip from tenants pool
        """
        try:
            # NOTE: the docs show this (in newer openstack?) as:
            # http://developer.openstack.org/api-ref-compute-v2-ext.html#DeleteFloatingIPBulk
            # data = {"ip_range":ip_range}
            data = {"floating_ips_bulk_delete": ip_range}
            server_resp = self.connection.request(
                '/os-floating-ips-bulk/delete',
                method='POST', data=data)
            return server_resp.object
        except Exception, e:
            raise

    def ex_bulk_create_floating_ips(self, ip_range, pool="nova", interface="eth0"):
        """
        Deallocate an existing floating_ip from tenants pool
        """
        try:
            data = {"floating_ips_bulk_create": {
                "ip_range": ip_range,
                "pool": pool,
                "interface":interface
                }}
            server_resp = self.connection.request(
                '/os-floating-ips-bulk',
                method='POST', data=data)
            return server_resp.object
        except Exception, e:
            raise

    def ex_add_fixed_ip(self, server, network_id):
        """
        """
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'addFixedIp': {'networkId': network_id}})
            return server_resp.object
        except Exception, e:
            raise

    def ex_remove_fixed_ip(self, server, fixed_ip_addr):
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'removeFixedIp': {'address': fixed_ip_addr}})
            return server_resp.object
        except Exception, e:
            raise

    def ex_associate_floating_ip(self, server, address, **kwargs):
        """
        Associate an allocated floating IP to the node
        """
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'addFloatingIp': {'address': address}})
            return server_resp.object
        except Exception, e:
            raise

    def ex_disassociate_floating_ip(self, server, address=None, **kwargs):
        """
        Disassociate a floating IP that's been associated to the node
        """
        try:
            if not address:
                public_ips = server._node.public_ips
                if not public_ips:
                    logger.warn("Could not determine public IP address,\
                    please provide the floating IP address")
                    return None
                address = public_ips[0]
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'removeFloatingIp': {'address': address}})
            return server_resp.object
        except Exception, e:
            raise

    #Security Groups
    def ex_add_security_group(self, server, sec_group, **kwargs):
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'addSecurityGroup': {'name': sec_group.name}})
            return server_resp.object
        except Exception, e:
            raise

    def ex_remove_security_group(self, server, sec_group):
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'removeSecurityGroup': {'name': sec_group.name}})
            return server_resp.object
        except Exception, e:
            raise


    #API Limits
    def ex_get_limits(self):
        """
        _to_rate and _to_absolute
        """
        def _to_rate(el):
            rate_limits = el.get('limits', {}).get('rate', [])
            limit_dict = {}
            for a_limit in rate_limits:
                rest_dict = {}
                for rest_limit in a_limit['limit']:
                    r_limit_map = {}
                    r_limit_map['call_available'] = rest_limit['value']
                    r_limit_map['call_limit'] = rest_limit['remaining']
                    r_limit_map['duration'] = rest_limit['unit']
                    r_limit_map['limit_expiry'] = rest_limit['next-available']
                    rest_dict[rest_limit['verb']] = r_limit_map
                limit_dict[a_limit['uri']] = rest_dict
            return limit_dict

        def _to_absolute(el):
            return el.get('limits', {}).get('absolute', {})

        json_limits = self.connection.request("/limits").object
        rate = _to_rate(json_limits)
        absolute = _to_absolute(json_limits)
        return {"rate": rate, "absolute": absolute}

    def ex_detach_interface(self, instance_id, port_id):
        """
        Detaches an existing fixed IP address (port) with ta server (Device)
        See:
            http://docs.openstack.org/api/openstack-compute/2/content/ext-os-interface.html
        """
        uri = "/servers/%s/os-interface/%s" % (instance_id, port_id)
        server_resp = self.connection.request(uri, method="DELETE")
        return server_resp.status == 202


    def ex_attach_interface(self, instance_id, network_id=None, port_id=None):
        """
        Attaches an existing fixed IP address (port) with ta server (Device)
        See:
            http://docs.openstack.org/api/openstack-compute/2/content/ext-os-interface.html
        """
        uri = "/servers/%s/os-interface" % (instance_id,)
        if not network_id and not port_id:
            raise Exception("Missing required argument: port_id OR network_id")
        elif network_id and port_id:
            raise Exception("Too many arguments: port_id OR network_id")
        elif port_id:
            attach_data = {"port_id":port_id }
        elif network_id:
            attach_data = {"net_id":network_id }

        server_resp = self.connection.request(uri, method="POST",
                data={"interfaceAttachment": attach_data})
        return server_resp.object['interfaceAttachment']

    def ex_os_services(self):
        """
        Return a list of services with their current state and status.

        See: http://docs.openstack.org/api\
             /openstack-compute/2/content/ext-os-services.html
        """
        uri = "/os-services"
        return self.connection.request(uri, method="GET").object["services"]

    """
    Private methods
    While these methods are useful,
    they will NOT be included when we push back to libcloud..
    """
    def neutron_associate_ip(self, node, *args, **kwargs):
        """
        Add IP (Neutron)
        There is no good way to interface libcloud + nova + neutron,
        instead we call neutronclient directly..
        Feel free to replace when a better mechanism comes along..
        """

        try:
            network_manager = self.get_network_manager()
            floating_ip = network_manager.associate_floating_ip(node.id)
        except NeutronClientException as q_error:
            if q_error.status_code == 409:
                #409 == Conflict
                #Lets look through the message and determine why:
                logger.info("Conflict stopped node from associating new "
                            "floating IP. Message=%s" % q_error.message)
            #Handle any conflicts that make sense and return, all others:
            raise

        return floating_ip

    def neutron_list_ips(self, node, *args, **kwargs):
        """
        List IP (Neutron)
        There is no good way to interface libcloud + nova + neutron,
        instead we call neutronclient directly..
        Feel free to replace when a better mechanism comes along..
        """

        try:
            network_manager = self.get_network_manager()
            floating_ips = network_manager.list_floating_ips()
        except NeutronClientException as q_error:
            if q_error.status_code == 409:
                #409 == Conflict
                #Lets look through the message and determine why:
                logger.info("Conflict stopped node from associating new "
                            "floating IP. Message=%s" % q_error.message)
            #Handle any conflicts that make sense and return, all others:
            raise
        ip_list = []
        for f_ip in floating_ips:
            if f_ip.get('instance_id') == node.id:
                ip_list.append(f_ip)
        return ip_list

    def get_user_manager(self):
        return UserManager.lc_driver_init(self)

    def get_network_manager(self):
        return NetworkManager.lc_driver_init(self)

    def neutron_disassociate_ip(self, node, *args, **kwargs):
        """
        Remove IP (Neutron)
        There is no good way to interface libcloud + nova + neutron,
        instead we call neutronclient directly..
        Feel free to replace when a better mechanism comes along..
        """

        try:
            network_manager = self.get_network_manager()
            network_manager.disassociate_floating_ip(node.id)
        except NeutronClientException as q_error:
            if q_error.status_code == 409:
                #409 == Conflict
                #Lets look through the message and determine why:
                logger.info("Conflict stopped node from disassociating new "
                            "floating IP. Message=%s" % q_error.message)
            #Handle any conflicts that make sense and return, all others:
            raise

        return True

    def _image_size(self, image):
        byte_size = image.extra['api']['OS-EXT-IMG-SIZE:size']
        in_gb = byte_size/1024**3
        return in_gb

    def ex_delete_ports(self, node, *args, **kwargs):
        """
        Delete Ports related to node. (Neutron)
        There is no good way to interface libcloud + nova + neutron,
        instead we use neutronclient directly..
        Hopefully Openstack provides a better option soon.
        """
        network_manager = NetworkManager.lc_driver_init(self)
        ports = network_manager.find_server_ports(node.id)
        for p in ports:
            network_manager.delete_port(p)

    # Metadata
    def ex_write_metadata(self, node, metadata, replace_metadata=True):
        """
        NOTE: Similar to ex_set_metadata, but allows for the option
        to KEEP existing metadata by setting 'replace_metadata=False'
        """
        #NOTE: PUT will REPLACE metadata each time it is added
        #      while POST will keep metadata that does not match
        #      The default for libcloud is to replace/override tags.
        # Ex:
        #     {'name': 'test_name'} + PUT {'tags': 'test_tag'}
        #     = {'tags': 'test_tag'}
        #     {'name': 'test_name'} + POST {'tags': 'test_tag'}
        #     = {'name': 'test_name', 'tags': 'test_tag'}
        #
        #
        method = 'PUT' if replace_metadata else 'POST'
        return self.connection.request(
            '/servers/%s/metadata' % (node.id,), method=method,
            data={'metadata': metadata}
        ).object['metadata']


    def ex_get_metadata(self, node, key=None):
        """
        Get a Node's metadata.

        @param      node: Node
        @type       node: L{Node}

        @param      key: Key associated with node's metadata.
        @type       node: L{str}

        @return: Key/Value metadata associated with node.
        @rtype: C{dict}
        """
        response = self.connection.request(
            '/servers/%s/metadata%s' % (node.id, '/%s' if key else ''),
            method='GET',)
        metadata = response.object['metadata']
        return metadata

    def ex_delete_metadata(self, node, key):
        """
        Sets the Node's metadata for a key.

        @param      node: Node
        @type       node: L{Node}

        @param      key: Key associated with node's metadata.
        @type       node: L{str}

        @rtype: C{bool}
        """
        resp = self.connection.request(
            '/servers/%s/metadata/%s' % (node.id, key,),
            method='DELETE')
        return resp.status == httplib.NO_CONTENT

    def ex_get_image_metadata(self, image, key):
        """
        Get an Image's metadata.

        @param      image: Image
        @type       image: L{Image}

        @param      key: Key associated with node's metadata.
        @type       node: L{str}

        @return: Key/Value metadata associated with an image.
        @rtype: C{dict}
        """
        response = self.connection.request(
            '/images/%s/metadata/%s' % (image.id, key,),
            method='GET',)
        metadata = response.object['metadata']
        return metadata

    def ex_get_image_metadata(self, image):
        """
        Get an Image's metadata.

        @param      image: Image
        @type       image: L{Image}

        @return: Key/Value metadata associated with an image.
        @rtype: C{dict}
        """
        response = self.connection.request(
            '/images/%s/metadata' % (image.id,),
            method='GET',)
        metadata = response.object['metadata']
        return metadata

    def ex_set_image_metadata(self, image, metadata):
        """
        Sets the Image's metadata.

        @param      image: Image
        @type       image: L{Image}

        @param      metadata: Key/Value metadata to associate with an image
        @type       metadata: C{dict}

        @rtype: C{dict}
        """
        return self.connection.request(
            '/images/%s/metadata' % (image.id,), method='POST',
            data={'metadata': metadata}
        ).object['metadata']

    def ex_replace_image_metadata(self, image, metadata):
        """
        Sets the Image's metadata.

        @param      image: Image
        @type       image: L{Image}

        @param      metadata: Key/Value metadata to associate with an image
        @type       metadata: C{dict}

        @rtype: C{dict}
        """
        self._json_safe_meta_values(metadata)
        return self.connection.request(
            '/images/%s/metadata' % (image.id,), method='PUT',
            data={'metadata': metadata}
        ).object['metadata']

    def ex_delete_image_metadata(self, image, key):
        """
        Deletes the Image's metadata for a key.

        @param      node: Image
        @type       node: L{Image}

        @param      key: Key associated with image's metadata.
        @type       node: L{str}

        @rtype: C{bool}
        """
        resp = self.connection.request(
            '/images/%s/metadata/%s' % (image.id, key,),
            method='DELETE')
        return resp.status == httplib.NO_CONTENT

    #Server Shelve Actions
    def ex_shelve_instance(self, server, **kwargs):
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'shelve': None})
            return server_resp.object
        except Exception, e:
            raise

    def ex_unshelve_instance(self, server):
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'unshelve': None})
            return server_resp.object
        except Exception, e:
            raise

    def ex_shelve_offload_instance(self, server):
        try:
            server_resp = self.connection.request(
                '/servers/%s/action' % server.id,
                method='POST',
                data={'shelveOffload': None})
            return server_resp.object
        except Exception, e:
            raise

    def ex_list_quota_for_user(self, user_id, tenant_id):
       """
       Shows quota for user and tenant_id combinations
       @keyword user_id: User ID to update. Typically a UUID.
       @type    user_id: C{str}
       @keyword tenant_id: Tenant or Project ID to update. Typically a UUID.
       @type    tenant_id: C{str}
       """
       server_resp = self.connection.request('/os-quota-sets/%s?user_id=%s'
                                             % (tenant_id, user_id))
       try:
           quota_obj = server_resp.object
           return (server_resp.status == 200, quota_obj)
       except Exception, e:
           logger.exception("Exception occured updating quota. Body:%s"
                            % body)
           return (False, None)

    def ex_update_quota(self, tenant_id, values):
       """
       Updates value/values in quota set
       @keyword tenant_id: Tenant or Project ID to update. Typically a UUID.
       @type    tenant_id: C{str}
       @keyword values: A Dict containing the new key/value for quota set
       @type    values: C{dict}
       """
       values['tenant_id'] = tenant_id
       body = {'quota_set': values}
       server_resp = self.connection.request('/os-quota-sets/%s'
                                             % (tenant_id,),
                                             method='PUT',
                                             data=body)
       try:
           quota_obj = server_resp.object
           return (server_resp.status == 200, quota_obj)
       except Exception, e:
           logger.exception("Exception occured updating quota. Body:%s"
                            % body)
           return (False, None)


    def ex_update_quota_for_user(self, tenant_id, user_id, values):
       """
       Updates value/values in quota set
       @keyword tenant_id: Tenant or Project ID to update. Typically a UUID.
       @type    tenant_id: C{str}
       @keyword user_id: User ID to update. Typically a UUID.
       @type    user_id: C{str}
       @keyword values: A Dict containing the new key/value for quota set
       @type    values: C{dict}
       """
       values['tenant_id'] = tenant_id
       body = {'quota_set': values}
       server_resp = self.connection.request('/os-quota-sets/%s?user_id=%s'
                                             % (tenant_id, user_id),
                                             method='PUT',
                                             data=body)
       try:
           quota_obj = server_resp.object
           return (server_resp.status == 200, quota_obj)
       except BaseHTTPError, e:
           if 'Quota limit' in e.msg or 'must be less' in e.msg or 'must less' in e.msg:
               return self.ex_update_quota(self, tenant_id, values)
       except Exception, e:
           logger.exception("Exception occured updating quota. Body:%s"
                            % body)
           return (False, None)




"""
Atmosphere service exceptions.

"""
from neutronclient.common.exceptions import NeutronClientException, NotFound as NeutronNotFound, BadRequest as NeutronBadRequest, NeutronException
from novaclient.exceptions import OverLimit as NovaOverLimit
try:
    from keystoneclient.exceptions import Unauthorized as KeystoneUnauthorized
except ImportError:
    from keystoneclient.apiclient.exceptions import Unauthorized as KeystoneUnauthorized

from glanceclient.exc import HTTPConflict as GlanceConflict, HTTPForbidden as GlanceForbidden, ClientException as GlanceClientException
from libcloud.common.types import InvalidCredsError as LibcloudInvalidCredsError, MalformedResponseError as LibcloudBadResponseError
from libcloud.compute.types import DeploymentError as LibcloudDeploymentError
from libcloud.common.exceptions import BaseHTTPError  as LibcloudHTTPError


class NonZeroDeploymentException(Exception):
    pass


class ServiceException(Exception):
    pass


class ConnectionFailure(ServiceException):
    pass

class MissingArgsException(ServiceException):
    pass



import os
import sys
import doctest

from distutils.core import Command
from unittest import TextTestRunner, TestLoader
from glob import glob
from os.path import splitext, basename, join as pjoin
import setuptools
from rtwo.version import get_version, read_requirements

readme = open('README.md').read()
dependencies, requirements = read_requirements('requirements.txt')

long_description = """
rtwo %s
A unified interface into multiple cloud providers.

To install use pip install git+git://git@github.com:iPlantCollaborativeOpenSource/rtwo.git

----

%s

----

For more information, please see: https://github.com/iPlantCollaborativeOpenSource/rtwo
""" % (get_version('short'), readme)


TEST_PATHS = ['rtwo/test',]
class TestCommand(Command):
    description = "run test suite"
    user_options = []

    def initialize_options(self):
        THIS_DIR = os.path.abspath(os.path.split(__file__)[0])
        sys.path.insert(0, THIS_DIR)
        for test_path in TEST_PATHS:
            sys.path.insert(0, pjoin(THIS_DIR, test_path))
        self._dir = os.getcwd()

    def finalize_options(self):
        pass

    def run(self):
        try:
            import mock
            mock
        except ImportError:
            print('Missing "mock" library. mock is library is needed '
                  'to run the tests. You can install it using pip: '
                  'pip install mock')
            sys.exit(1)


        status = self._run_tests()
        sys.exit(status)

    def _run_tests(self):
        secrets_current = pjoin(self._dir, 'rtwo/test', 'secrets.py')
        secrets_dist = pjoin(self._dir, 'rtwo/test', 'secrets.py.dist')

        if not os.path.isfile(secrets_current):
            print("Missing " + secrets_current)
            print("Maybe you forgot to copy it from .dist:")
            print("cp rtwo/test/secrets.py.dist rtwo/test/secrets.py")
            sys.exit(1)

        mtime_current = os.path.getmtime(secrets_current)
        mtime_dist = os.path.getmtime(secrets_dist)

        if mtime_dist > mtime_current:
            print("It looks like test/secrets.py file is out of date.")
            print("Please copy the new secrets.py.dist file over otherwise" +
                  " tests might fail")

        testfiles = []
        for test_path in TEST_PATHS:
            for t in glob(pjoin(self._dir, test_path, 'test_*.py')):
                testfiles.append('.'.join(
                    [test_path.replace('/', '.'), splitext(basename(t))[0]]))
        tests = TestLoader().loadTestsFromNames(testfiles)

        t = TextTestRunner(verbosity=2)
        res = t.run(tests)
        return not res.wasSuccessful()

setuptools.setup(
    name='rtwo',
    version=get_version('short'),
    author='iPlant Collaborative',
    author_email='atmodevs@gmail.com',
    description="A unified interface into multiple cloud providers.",
    long_description=long_description,
    license="BSD License, 3 clause",
    url="https://github.com/iPlantCollaborativeOpenSource/rtwo",
    packages=setuptools.find_packages(),
    dependency_links=dependencies,
    install_requires=requirements,
    cmdclass={
        'test': TestCommand,
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries",
        "Topic :: System",
        "Topic :: System :: Clustering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Systems Administration"
    ])



"""
Common functions used by all Openstack managers.
"""
import copy
import sys

import glanceclient
from keystoneclient.exceptions import AuthorizationFailure
from keystoneclient import exceptions
from swiftclient import client as swift_client
from novaclient import client as nova_client
from novaclient import api_versions
from neutronclient.v2_0 import client as neutron_client
from openstack import connection as openstack_sdk
from keystoneauth1.identity import v3
from keystoneauth1.session import Session
from libcloud.compute.deployment import ScriptDeployment

from threepio import logger

from rtwo.version import version_str as rtwo_version


class LoggedScriptDeployment(ScriptDeployment):

    def __init__(self, script, name=None, delete=False, logfile=None):
        """
        Use this for client-side logging
        """
        super(LoggedScriptDeployment, self).__init__(
            script, name=name, delete=delete)
        if logfile:
            self.script = self.script + " &> %s" % logfile
        #logger.info(self.script)

    def run(self, node, client):
        """
        Server-side logging
        """
        node = super(LoggedScriptDeployment, self).run(node, client)
        if self.stdout:
            logger.debug('%s (%s)STDOUT: %s' % (node.id, self.name,
                                                self.stdout))
        if self.stderr:
            logger.warn('%s (%s)STDERR: %s' % (node.id, self.name,
                                               self.stderr))
        return node


def _connect_to_swift(*args, **kwargs):
    """
    """
    swift = swift_client.Connection(*args, **kwargs)
    return swift


def _connect_to_neutron(*args, **kwargs):
    """
    """
    if 'auth_url' in kwargs and 'v2.0' not in kwargs['auth_url']:
        kwargs['auth_url'] += "/v2.0"
    neutron = neutron_client.Client(*args, **kwargs)
    neutron.format = 'json'
    return neutron


def _connect_to_keystone_v3(
        auth_url, username, password,
        project_name, domain_name="default"):
    """
    Given a username and password,
    authenticate with keystone to get an unscoped token
    Exchange token to receive an auth,session,token scoped to a specific project_name and domain_name.
    """
    unscoped_auth = v3.Password(
        username=username, password=password,
        auth_url=auth_url, user_domain_name=domain_name,
        unscoped=True)
    unscoped_sess = Session(auth=unscoped_auth)
    unscoped_token = unscoped_sess.get_token()
    return _token_to_keystone_scoped_project(auth_url, unscoped_token, project_name, domain_name)

def _token_to_keystone_scoped_project(
        auth_url, token,
        project_name, domain_name="default"):
    """
    Given an auth_url and scoped/unscoped token:
    Create an auth,session and token for a specific project_name and domain_name (Required to access a serviceCatalog for neutron/nova/etc!)
    """
    auth = v3.Token(auth_url=auth_url, token=token, project_name=project_name, project_domain_id=domain_name)
    sess = Session(auth=auth)
    token = sess.get_token()
    return (auth, sess, token)



def _connect_to_keystone(*args, **kwargs):
    """
    """
    try:
        version = kwargs.get('version', 'v2.0')
        if version == 'v2.0':
            from keystoneclient.v2_0 import client as ks_client
        else:
            from keystoneclient.v3 import client as ks_client
        keystone = ks_client.Client(*args, **kwargs)
    except AuthorizationFailure as e:
        raise Exception("Authorization Failure: Bad keystone secrets or "
                        "firewall causing a timeout.")
    return keystone


def _connect_to_openstack_sdk(*args, **kwargs):
    """
    Connect to OpenStack SDK client
    """

    # Atmosphere was configured on 'v2' naming.
    # This will update the value to the current naming, 'project_name'
    from openstack import profile
    from openstack import utils
    identity_version = kwargs.get('identity_api_version', 2)
    if identity_version == 2:
        return None
    utils.enable_logging(True, stream=sys.stdout) # TODO: stream this to _not_ stdout
    user_profile = profile.Profile()
    user_profile.set_region(profile.Profile.ALL, kwargs.get('region_name'))
    if 'project_name' not in kwargs and 'tenant_name' in kwargs:
        kwargs['project_name'] = kwargs.pop('tenant_name')

    user_profile.set_version('identity', 'v%s' % identity_version)
    user_profile.set_interface('identity', 'admin')
    user_agent = "rtwo/%s" % (rtwo_version(),)
    stack_sdk = openstack_sdk.Connection(
        user_agent=user_agent,
        profile=user_profile,
        **kwargs
    )
    return stack_sdk

def _connect_to_glance(keystone, version='1', *args, **kwargs):
    """
    NOTE: We use v1 because moving up to v2 results in a LOSS OF
    FUNCTIONALITY..
    """
    glance_endpoint = keystone.service_catalog.url_for(
        service_type='image',
        endpoint_type='publicURL')
    auth_token = keystone.service_catalog.get_token()
    if type(version) == str:
        if '3' in version:
            version = 2
        elif '2' in version:
            version = 2
    glance = glanceclient.Client(version,
                                 endpoint=glance_endpoint,
                                 token=auth_token['id'])
    return glance

def _connect_to_keystoneauth(
            auth_url, username, password,
            user_domain_id, project_domain_id):
    """
    Connect to keystoneauth (Password) - The v3 way
    ..Because obviously.. So simple.
    """
    keystone_auth = v3.Password(
        auth_url, username=username,
        password=password, user_domain_id=user_domain_id,
        project_domain_id=project_domain_id)
    return keystone_auth

def _connect_to_nova(*args, **kwargs):
    kwargs = copy.deepcopy(kwargs)
    version = kwargs.pop('version', '2')
    auth_url = kwargs.pop('auth_url')
    if auth_url.endswith('/'):
        auth_url = auth_url[:-1]
    username = kwargs.pop('username')
    password = kwargs.pop('password')
    tenant_name = kwargs.pop('tenant_name')
    region_name = kwargs.pop('region_name')
    project_domain_id = kwargs.pop('project_domain_id','default')
    user_domain_id = kwargs.pop('user_domain_id','default')
    endpoint_type = kwargs.pop('endpoint_type','publicURL')
    service_type = kwargs.pop('service_type','compute')
    version = api_versions.APIVersion("2.0")
    password_auth = _connect_to_keystoneauth(
        auth_url, username, password,
        user_domain_id, project_domain_id)
    #kwargs['auth'] = password_auth
    kwargs['http_log_debug'] = True
    nova = nova_client.Client(version,
                              username,
                              password,
                              tenant_name,
                              auth_url=auth_url,
                              region_name=region_name,
                              #extensions = self.extensions,
                              *args, no_cache=True, **kwargs)
    return nova


def findall(manager, *args, **kwargs):
    """
        Find all items with attributes matching ``**kwargs``.

        This isn't very efficient: it loads the entire list then filters on
        the Python side.
    """
    found = []
    searches = kwargs.items()

    for obj in manager.list():
        try:
            if all(getattr(obj, attr) == value
                   for (attr, value) in searches):
                found.append(obj)
        except AttributeError:
            continue
    return found


def find(manager, **kwargs):
        """
        Find a single item with attributes matching ``**kwargs``.

        This isn't very efficient: it loads the entire list then filters on
        the Python side.
        """
        rl = findall(manager, **kwargs)
        num = len(rl)

        if num == 0:
            msg = "No %s matching %s." % (manager.resource_class.__name__,
                                          kwargs)
            raise exceptions.NotFound(404, msg)
        elif num > 1:
            raise exceptions.NoUniqueMatch
        else:
            return rl[0]



"""
UserManager:
    Remote Eucalyptus Admin controls..

User, BooleanField, StringList
    These XML parsing classes belong to euca_admin.py,
    and can be found on the Cloud Controller
"""
import sys
import boto
from boto.ec2.regioninfo import RegionInfo
from boto.exception import EC2ResponseError

from urlparse import urlparse
from threepio import logger

#Enter euca admin credentials here!
EUCA_ADMIN_KEY = ""
EUCA_ADMIN_SECRET = ""
EUCA_EC2_URL = ""


class UserManager():
    """
    Accessing Eucalyptus' Admin panel works
    just like any other EC2-compliant query
    HOWEVER, the default path for admin is
    '/services/Accounts/' instead of '/services/Eucalyptus'!
    """
    admin_key = None
    admin_secret = None
    admin_ec2_url = None
    account_path = None
    connection = None

    def __init__(self, *args, **kwargs):
        key = kwargs.get('key','')
        secret = kwargs.get('secret','')
        url = kwargs.get('url','')
        path = kwargs.get('account_path','/services/Accounts')

        self.admin_key = key
        self.admin_secret = secret
        self.admin_ec2_url = url
        self.account_path = path

        parsed = urlparse(url)
        region = RegionInfo(None, 'eucalyptus', parsed.hostname)
        self.connection = boto.connect_ec2(
            aws_access_key_id=self.admin_key,
            aws_secret_access_key=self.admin_secret,
            is_secure=False, region=region,
            port=parsed.port, path=self.account_path)
        self.connection.APIVersion = 'eucalyptus'

    def delete_user(self, username):
        """
        Deletes 'username' from eucalyptus,
        returns True on success
        """
        try:
            reply = self.connection.get_object(
                'DeleteUser', {'UserName': username}, BooleanResponse)
        except EC2ResponseError:
            logger.info("User does not exist")
            return False

        return reply and reply.reply == 'true'

    def add_user(self, username, email="N/A", adminUser=False):
        """
        Adds new user, with 'username' and 'email' to eucalyptus.
        returns True on success
        """
        reply = self.connection.get_object(
            'AddUser', {'UserName': username, 'Email': email,
                        'Admin': adminUser}, BooleanResponse)
        return reply and reply.reply == 'true'

    def get_keys(self, userList):
        """
        Returns a dictionary of keys for each username in userList
        """
        userDict = self.get_users(userList)
        user_keys_dict = {}
        for username in userDict.keys():
                userObj = userDict[username]
                user_keys_dict[username] = {
                    'username': username,
                    'access_key': userObj['access_key'],
                    'secret_key': userObj['secret_key']}
        return user_keys_dict

    def get_users(self, userList):
        """
        Returns a dictionary of username:attributes
        """
        params = {}
        self.connection.build_list_params(params, userList, 'UserNames')
        euca_list = self.connection.get_list(
            'DescribeUsers', params, [('euca:item', User)])
        return self.userListDict(euca_list)

    def userListDict(self, euca_list):
        """
        Utility method to turn a user_list
        (XML Parsed object from euca_admin) into python dict
        """
        userDict = {}
        for user in euca_list:
		logger.info(user.__dict__)
                userDict[user.user_userName] = {
                    'username': user.user_userName,
                    'access_key': user.user_accessKey,
                    'secret_key': user.user_secretKey,
                    'certificateSerial': user.user_certificateSerial,
                    'certificateCode': user.user_certificateCode,
                    'confirmationCode': user.user_confirmationCode,
                    'confirmed': user.user_confirmed,
                    'admin': user.user_admin,
                    'enabled': user.user_enabled,
                    'email': user.user_email,
                }
        return userDict

    #Useful calls you will want
    def get_key(self, username):
            return self.get_keys([username])[username]

    def get_all_users(self):
            return self.get_users([])

    def get_user(self, username):
            return self.get_users([username])[username]

#NOTE: I DO NOT OWN THE RIGHTS TO ANY OF THE CODE BELOW!
# The functions below are for used to parse the XML response
# from eucalyptus, and are placed here to avoid dependencies on euca_admin
##################


class User():
    """
    The user object stores the XML response for DescribeUsers
    This class was pulled from euca_admin/users.py
    """

    def __init__(self, userName=None, email="N/A", certificateCode=None,
                 confirmationCode=None, accessKey=None, secretKey=None,
                 confirmed=False, admin=False, enabled=False,
                 distinguishedName=None, certificateSerial=None):
        self.user_userName = userName
        self.user_email = email
        self.user_distinguishedName = distinguishedName
        self.user_certificateSerial = certificateSerial
        self.user_certificateCode = certificateCode
        self.user_confirmationCode = confirmationCode
        self.user_accessKey = accessKey
        self.user_secretKey = secretKey
        self.user_confirmed = confirmed
        self.user_admin = admin
        self.user_enabled = enabled
        self.user_groups = StringList()
        self.user_revoked = StringList()
        self.user_list = self.user_groups
        self.euca = None

    def __repr__(self):
        r = 'USER\t\t%s\t%s%s\t%s' \
            % (self.user_userName, self.user_email,
               '\tADMIN' if self.user_admin == 'true' else ' ',
               'ENABLED' if self.user_enabled == 'true' else 'DISABLED')
        r = '%s\nUSER-GROUP\t%s\t%s' \
            % (r, self.user_userName, self.user_groups)
        r = '%s\nUSER-CERT\t%s\t%s\t%s' \
            % (r, self.user_userName, self.user_distinguishedName,
               self.user_certificateSerial)
        r = '%s\nUSER-KEYS\t%s\t%s\t%s' \
            % (r, self.user_userName, self.user_accessKey, self.user_secretKey)
        r = '%s\nUSER-CODE\t%s\t%s' \
            % (r, self.user_userName, self.user_certificateCode)
        r = '%s\nUSER-WEB \t%s\t%s' \
            % (r, self.user_userName, self.user_confirmationCode)
        return r

    def startElement(self, name, attrs, connection):
        if name == 'euca:groups':
            return self.user_groups
        elif name == 'euca:revoked':
            return self.user_revoked
        else:
            return None

    def endElement(self, name, value, connection):
        if name == 'euca:userName':
            self.user_userName = value
        elif name == 'euca:email':
            self.user_email = value
        elif name == 'euca:admin':
            self.user_admin = value
        elif name == 'euca:confirmed':
            self.user_confirmed = value
        elif name == 'euca:enabled':
            self.user_enabled = value
        elif name == 'euca:distinguishedName':
            self.user_distinguishedName = value
        elif name == 'euca:certificateSerial':
            self.user_certificateSerial = value
        elif name == 'euca:certificateCode':
            self.user_certificateCode = value
        elif name == 'euca:confirmationCode':
            self.user_confirmationCode = value
        elif name == 'euca:accessKey':
            self.user_accessKey = value
        elif name == 'euca:secretKey':
            self.user_secretKey = value
        elif name == 'euca:entry':
            self.user_list.append(value)
        else:
            setattr(self, name, value)


class BooleanResponse:
    """
    XML Parsing class for add/delete user
    Pulled from euca_admin/generic.py
    """
    def __init__(self, reply=False):
        self.reply = reply
        self.error = None

    def __repr__(self):
        if self.error:
            print 'RESPONSE %s' % self.error
            sys.exit(1)
        else:
            return 'RESPONSE %s' % self.reply

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'euca:_return':
            self.reply = value
        elif name == 'Message':
            self.error = value
        else:
            setattr(self, name, value)


class StringList(list):
    """
    XML Parsing class for list of users in describeUser
    Pulled from euca_admin/generic.py
    """
    def __repr__(self):
        r = ""
        for i in self:
            r = "%s %s" % (r, i)
        return r

    def startElement(self, name, attrs, connection):
        pass

    def endElement(self, name, value, connection):
        if name == 'euca:entry':
            self.append(value)



"""
Atmosphere service volume.

"""
from abc import ABCMeta

from threepio import logger

from rtwo.models.provider import AWSProvider, EucaProvider, OSProvider


class BaseVolume(object):
    __metaclass__ = ABCMeta


class MockVolume(BaseVolume):

    def __init__(self, volume_id, provider):
        self._volume = None
        self.id = volume_id
        self.alias = volume_id
        self.size = -1
        self.attachment_set = []
        self.extra = {}
        self.name = "Mock Volume %s" % volume_id
        self.provider = provider


class Volume(BaseVolume):

    provider = None

    machine = None

    def __init__(self, lc_volume):
        self._volume = lc_volume
        self.id = lc_volume.id
        self.alias = lc_volume.id
        self.attachment_set = lc_volume.extra['attachments']
        self.extra = lc_volume.extra
        self.name = lc_volume.name
        self.provider = self.provider
        self.size = lc_volume.size

    @classmethod
    def get_volumes(cls, volumes):
        return map(cls.provider.volumeCls, volumes)

    # order matters with reset methods.
    def reset(self):
        Volume.reset()
        self._volume = None
        self.machine = None

    # again order matters... /sigh
    @classmethod
    def reset(cls):
        cls._volume = None
        cls.machine = None

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return reduce(lambda x, y: x+y,
                      map(unicode, [self.__class__, " ", self.json()]))

    def __repr__(self):
        return str(self)

    def json(self):
        return {'id': self.id,
                'alias': self.alias,
                'attachment_set': self.attachment_set,
                'extra': self.extra,
                'name': self.name,
                'provider': self.provider.name,
                'size': self.size}


class AWSVolume(Volume):

    provider = AWSProvider


class EucaVolume(Volume):

    provider = EucaProvider


class OSVolume(Volume):

    provider = OSProvider



"""
OpenStack Network Admin Libarary

    Create:
    # To create a libcloud driver use the lc_driver_init method.
    # For instance with an rtwo OpenStack driver it's driver._connection.
    nm = NetworkManager.lc_driver_init(driver._connection)

    Use this library to:
      * manage networks within Neutron - openstack networking
"""
import os
import netaddr


from threepio import logger

from rtwo.drivers.common import _connect_to_neutron
from neutronclient.common.exceptions import NeutronClientException, NotFound

ROUTER_INTERFACE_NAMESPACE = (
    'network:router_interface',
    'network:router_interface_distributed',
    'network:ha_router_replicated_interface'
)

class NetworkManager(object):

    neutron = None
    default_router = None

    def __init__(self, *args, **kwargs):
        self.default_router = kwargs.pop("router_name", None)
        self.neutron = self.new_connection(*args, **kwargs)

    def new_connection(self, *args, **kwargs):
        """
        Allows us to make another connection (As the user)
        """
        if hasattr(kwargs, 'auth_url'):
            #HACK - Nova is certified-broken-on-v3. 
            kwargs['version'] = 'v2.0'
            kwargs['auth_url'] = kwargs.get('auth_url','').replace('v3','v2.0')
        neutron = _connect_to_neutron(*args, **kwargs)
        return neutron

    def tenant_networks(self, tenant_id=None):
        if not tenant_id:
            tenant_id = self.get_tenant_id()
        tenant_nets = self.list_networks(tenant_id=tenant_id)
        return tenant_nets

    def get_tenant_id(self):
        credentials = self.get_credentials()
        try:
            tenant_id = credentials.get('auth_tenant_id')
            return tenant_id
        except KeyError:
            logger.warn(
                "Key 'auth_tenant_id' no longer exists in"
                "'get_credentials()'")
            return None

    def get_credentials(self):
        """
        Return the user_id and tenant_id of the network manager
        """
        auth_info = self.neutron.httpclient.get_auth_info()
        if not auth_info.get('auth_tenant_id'):
            self.list_networks()
            auth_info = self.neutron.httpclient.get_auth_info()
        auth_info.pop('auth_token')
        return auth_info

    ##Admin-specific methods##
    def project_network_map(self):
        named_networks = self.find_network('-net', contains=True)
        users_with_networks = [net['name'].replace('-net', '')
                               for net in named_networks]
        user_map = {}
        networks = self.list_networks()
        subnets = self.list_subnets()
        routers = self.list_routers()
        ports = self.list_ports()
        for user in users_with_networks:
            my_nets = [net for net in networks if '%s-net' % user in net['name']]
            net_ids = [n['id'] for n in my_nets]
            my_subnets = [subnet for subnet in subnets if '%s-subnet' % user in subnet['name']]
            subnet_ids = [s['id'] for s in my_subnets]
            my_ports = []
            for port in ports:
                if 'dhcp' in port['device_owner'] or \
                        'compute:None' in port['device_owner']:
                    #Skip these ports..
                    continue
                if port['network_id'] in net_ids:
                    my_ports.append(port)
                    continue
                fixed_ips = port['fixed_ips']
                for fixed_ip in fixed_ips:
                    if fixed_ip['subnet_id'] in subnet_ids:
                        my_ports.append(port)
                        break
            #TODO: Can you have more than one of these?
            if len(my_nets) == 1:
                my_nets = my_nets[0]
            if len(my_subnets) == 1:
                my_subnets = my_subnets[0]
            user_map[user] = {'network': my_nets,
                              'subnet': my_subnets,
                              'public_interface': my_ports}
            logger.debug("Added user %s" % user_map[user])
        return user_map

    def get_user_neutron(self, username, password,
                         project_name, auth_url, region_name):
        user_creds = {
            'username': username,
            'password': password,
            'tenant_name': project_name,
            'auth_url': auth_url,
            'region_name': region_name
        }
        user_neutron = self.new_connection(**user_creds)
        return user_neutron

    def disassociate_floating_ip(self, server_id):
        """
        Remove floating IP from the server <server_id>
        * NO return value
        * raises NeutronClientException if delete fails
        """
        floating_ip_id = None
        for f_ip in self.list_floating_ips():
            if f_ip.get('instance_id') == server_id:
                floating_ip_id = f_ip['id']
        #No floating ip matches - Disassociate has nothing to do
        if not floating_ip_id:
            return
        #Remove floating ip
        deleted_ip = self.neutron.delete_floatingip(floating_ip_id)
        return

    def associate_floating_ip(self, server_id):
        """
        Create a floating IP on the external network
        Find port of new VM
        Associate new floating IP with the port assigned to the new VM
        """
        networks = self.list_networks()
        external_networks = [net for net in networks 
                             if net['router:external'] == True]
        if not external_networks:
            raise Exception("CONFIGURATION ERROR! No external networks found!"
                            " Cannot associate floating ip without it!"
                            " Create a fixed IP/port first!")

        instance_ports = self.list_ports(device_id=server_id)
        if not instance_ports:
            raise Exception("No ports found with device_id == %s."
                            " Create a fixed IP/port first!" % server_id)
        #TODO: Look at the network if it already has a floating ip, dont
        #re-create
        body = {'floatingip': {
                   'port_id': instance_ports[0]['id'],
                   'floating_network_id': external_networks[0]['id']
               }}
        new_ip = self.neutron.create_floatingip(body)['floatingip']

        logger.info('Assigned Floating IP - %s:%s' % (server_id, new_ip))
        return new_ip

    def create_port(self, server_id, network_id, subnet_id=None,
            ip_address=None, tenant_id=None, mac_address=None, name=None):
        """
        Create a new (Fixed IP) Port between server id and the user network
        """
        if not name:
            name = 'fixed_ip_%s' % (server_id,)
        port_data = {'port':
                {
                    "tenant_id": tenant_id,
                    "network_id":network_id,
                    "device_id":server_id,
                    "fixed_ips": [{"subnet_id":subnet_id, "ip_address":
                        ip_address}],
                    'admin_state_up':True,
                    'name':name
                }
            }
        if mac_address:
            port_data['port']['mac_address'] = mac_address
        if subnet_id and ip_address:
            #In this case, we should attach the interface after the fact.
            port_data['port'].pop('device_id')
        port_obj = self.neutron.create_port(port_data)
        return port_obj['port']


    def find_server_ports(self, server_id):
        """
        Find all the ports for a given server_id (device_id in port object).
        """
        server_ports = []
        all_ports = self.list_ports()
        return [p for p in all_ports if p['device_id'] == server_id]

    def list_floating_ips(self):
        instance_ports = self.list_ports()
        floating_ips = self.neutron.list_floatingips()['floatingips']
        # Connect instances and floating_ips using ports.
        for fip in floating_ips:
            port = filter(lambda(p): p['id'] == fip['port_id'], instance_ports)
            if port:
                fip['instance_id'] = port[0]['device_id']
        #logger.debug(floating_ips)
        return floating_ips

    def rename_security_group(self, project, security_group_name=None):
        security_group_resp = self.neutron.list_security_groups(
                tenant_id=project.id)
        default_group_id = None
        for sec_group in security_group_resp['security_groups']:
            if 'default' in sec_group['name']:
                default_group_id = sec_group['id']
                break
        if not default_group_id:
            raise Exception("Could not find the security group named 'default'")
        try:
            if not security_group_name:
                security_group_name = project.name
            #FIXME: we don't actually name it?
            sec_group = self.neutron.update_security_group(
                    default_group_id,
                    {"security_group": {"description": security_group_name}})
            return sec_group
        except NeutronClientException:
            logger.exception("Problem updating description of 'default'"
                             "security group to %s" % project.name)
            raise


    ##Libcloud-Neutron Interface##
    @classmethod
    def lc_driver_init(self, lc_driver, *args, **kwargs):
        lc_driver_args = {
            'username': lc_driver.key,
            'password': lc_driver.secret,
            'tenant_name': lc_driver._ex_tenant_name,
            #Libcloud requires /v2.0/tokens -- OS clients do not.
            'auth_url': lc_driver._ex_force_auth_url.replace('/tokens',''),
            'region_name': lc_driver._ex_force_service_region}
        lc_driver_args.update(kwargs)
        manager = NetworkManager(*args, **lc_driver_args)
        return manager

    def lc_list_networks(self, *args, **kwargs):
        """
        Call neutron list networks and convert to libcloud objects
        """
        network_list = self.neutron.list_networks(*args, **kwargs)
        return [self._to_lc_network(net) for net in network_list['networks']]

    def _to_lc_network(self, net):
        from libcloud.compute.drivers.openstack import OpenStackNetwork
        return OpenStackNetwork(id=net['id'],
                                name=net['name'],
                                cidr=net.get('cidr', None),
                                extra=net,
                                driver=self)
    ##GET##
    def get_network(self, network_id):
        for net in self.neutron.list_networks()['networks']:
            if network_id == net['id']:
                return net
        return None

    def get_subnet(self, subnet_id):
        for subnet in self.neutron.list_subnets()['subnets']:
            if subnet_id == subnet['id']:
                return subnet
        return None

    def get_port(self, port_id):
        ports = self.list_ports()
        if not ports:
            return []
        for port in ports:
            if port['id'] == port_id:
                return port
        return None
    ##Easy Lists##
    def list_networks(self, *args, **kwargs):
        """
        NOTE: kwargs can be: tenant_id=, or any other attr listed in the
        details of a network.
        """
        return self.neutron.list_networks(*args, **kwargs)['networks']

    def list_subnets(self):
        return self.neutron.list_subnets()['subnets']

    def list_routers(self):
        return self.neutron.list_routers()['routers']

    ##LOOKUP##
    def find_tenant_resources(self, tenant_id, instance_ids=[]):
        networks = [net for net in self.list_networks()
                    if net['tenant_id'] == tenant_id]
        ports = [port for port in self.list_ports()
                 if port['tenant_id'] == tenant_id
                 or port['device_id'] in instance_ids]
        subnets = [subnet for subnet in self.list_subnets()
                    if subnet['tenant_id'] == tenant_id]
        routers = [router for router in self.list_routers()
                    if router['tenant_id'] == tenant_id]
        return {"ports": ports,
                "networks": networks,
                "subnets": subnets,
                "routers": routers
               }
    def find_network(self, network_name, contains=False):
        return [net for net in self.list_networks()
                if network_name == net['name']
                or (contains and network_name in net['name'])]

    def find_subnet(self, subnet_name, contains=False):
        return [net for net in self.list_subnets()
                if subnet_name == net['name']
                or (contains and subnet_name in net['name'])]

    def find_router(self, router_name):
        return [net for net in self.list_routers()
                if router_name == net['name']]

    def find_ports_for_router(self, router_name):
        routers = self.find_router(router_name)
        if not routers:
            return []
        router_id = routers[0]['id']
        return [port for port in self.list_ports()
                if router_id == port['device_id']]

    def list_ports(self, **kwargs):
        """
        Options:
        subnet_id=subnet.id
        device_id=device.id
        ip_address=111.222.333.444
        """
        return self.neutron.list_ports(**kwargs)['ports']

    def find_router_interface(self, router, subnet):
        #If no router/subnet, return None
        if not router or not subnet:
            return None
        #If str router/subnet, find the obj
        if type(router) != dict:
            routers = self.find_router(router)
            if not routers:
                logger.info('Router %s does not exists' % router)
                return None
            router = routers[0]

        if type(subnet) != dict:
            subnets = self.find_subnet(subnet)
            if not subnets:
                logger.info('Subnet %s does not exists' % subnet)
                return None
            subnet = subnets[0]

        #Return the router interfaces matching router+subnet
        router_name = router['name']
        subnet_id = subnet['id']
        router_ports = self.find_ports_for_router(router_name)
        router_interfaces = []
        for port in router_ports:
            if port['device_owner'] not in ROUTER_INTERFACE_NAMESPACE:
                continue
            subnet_match = False
            for ip_subnet_obj in port['fixed_ips']:
                if subnet_id in ip_subnet_obj['subnet_id']:
                    subnet_match = True
                    break
            if subnet_match:
                router_interfaces.append(port)
        return router_interfaces

    def find_router_gateway(self,
                            router_name,
                            external_network_name='ext_net'):
        network_id = self.find_network(external_network_name)[0]['id']
        routers = self.find_router(router_name)
        if not routers:
            return
        return [r for r in routers if r.get('external_gateway_info') and
                network_id in r['external_gateway_info'].get('network_id', '')]

    ##ADD##
    def create_network(self, neutron, network_name):
        existing_networks = self.find_network(network_name)
        if existing_networks:
            logger.info('Network %s already exists' % network_name)
            return existing_networks[0]

        network = {'name': network_name, 'admin_state_up': True}
        network_obj = neutron.create_network({'network': network})
        return network_obj['network']


    def validate_cidr(self, cidr):
        logger.info("Attempting to validate cidr %s" % cidr)
        test_cidr_set = netaddr.IPSet([cidr])
        all_subnets = [subnet for subnet in self.list_subnets()
                       if subnet.get('ip_version', 4) != 6]
        all_subnet_ips = [sn['allocation_pools'] for sn in all_subnets]
        for idx, subnet_ip_list in enumerate(all_subnet_ips):
            for subnet_ip_range in subnet_ip_list:
                (start, end) = (subnet_ip_range['start'], subnet_ip_range['end'])
                if start.startswith('10') or end.startswith('10') or start.startswith('192') or end.startswith('192'):
                    continue
                test_range = netaddr.IPRange(
                    subnet_ip_range['start'], subnet_ip_range['end'])
                if len(test_range) > 1000:
                    continue
                for ip in test_range:
                    if ip in test_cidr_set:
                        raise Exception("Overlap detected for CIDR %s and Subnet %s" % (cidr, all_subnets[idx]))
        return True

    def create_subnet(self, neutron, subnet_name,
                      network_id, ip_version=4, cidr='172.16.1.0/24',
                      dns_nameservers=[]):
        existing_subnets = self.find_subnet(subnet_name)
        if existing_subnets:
            logger.info('Subnet %s already exists' % subnet_name)
            return existing_subnets[0]
        #self.validate_cidr(cidr)
        if not dns_nameservers:
            dns_nameservers = ['8.8.8.8', '8.8.4.4']
        subnet = {
            'name': subnet_name,
            'network_id': network_id,
            'ip_version': ip_version,
            'cidr': cidr,
            'dns_nameservers': dns_nameservers
        }
        logger.debug(subnet)
        subnet_obj = neutron.create_subnet({'subnet': subnet})
        return subnet_obj['subnet']

    def create_router(self, neutron, router_name):
        existing_routers = self.find_router(router_name)
        if existing_routers:
            logger.info('Router %s already exists' % router_name)
            return existing_routers[0]
        router = {'name': router_name, 'admin_state_up': True}
        router_obj = neutron.create_router({'router': router})
        return router_obj['router']

    def add_router_interface(self, router, subnet, interface_name=None):
        existing_router_interfaces = self.find_router_interface(router, subnet)
        if existing_router_interfaces:
            logger.info('Router Interface for Subnet:%s-Router:%s already'
                        'exists' % (subnet['name'], router['name']))
            return existing_router_interfaces[0]
        body = {"subnet_id": subnet['id']}
        interface_obj = self.neutron.add_interface_router(router['id'], body)
        if interface_name:
            self.neutron.update_port(
                    interface_obj['port_id'],
                    {"port":{"name":interface_name}})
        return interface_obj

    def set_router_gateway(self, neutron, router_name,
                           external_network_name='ext_net'):
        """
        Must be run as admin
        """
        existing_gateways = self.find_router_gateway(router_name,
                                                     external_network_name)
        if existing_gateways:
            logger.info('Router gateway for External Network:%s-Router:%s\
                already exists' % (external_network_name, router_name))
            return existing_gateways[0]
        #Establish the router_gateway
        router_id = self.get_router_id(self.neutron, router_name)
        external_network = self.get_network_id(neutron, external_network_name)
        body = {'network_id': external_network}
        return self.neutron.add_gateway_router(router_id, body)

    ## LOOKUPS##
    def get_subnet_id(self, neutron, subnet_name):
        sn_list = neutron.list_subnets(name=subnet_name)
        if sn_list and sn_list.get('subnets'):
            return sn_list['subnets'][0]['id']

    def get_router_id(self, neutron, router_name):
        rt_list = neutron.list_routers(name=router_name)
        if rt_list and rt_list.get('routers'):
            return rt_list['routers'][0]['id']

    def get_network_id(self, neutron, network_name):
        nw_list = neutron.list_networks(name=network_name)
        if nw_list and nw_list.get('networks'):
            return nw_list['networks'][0]['id']

    ##DELETE##
    def remove_router_gateway(self, router_name):
        router_id = self.get_router_id(self.neutron, router_name)
        if router_id:
            return self.neutron.remove_gateway_router(router_id)

    def remove_router_interface(self, neutron, router_name, subnet_name):
        router_id = self.get_router_id(self.neutron, router_name)
        subnet_id = self.get_subnet_id(neutron, subnet_name)
        #FIXME: Ensure no instances/IPs are using the interface
        # && raise an error if they try!
        if router_id and subnet_id:
            try:
                return neutron\
                    .remove_interface_router(router_id,
                                             {"subnet_id": subnet_id})
            except NeutronClientException, neutron_err:
                if 'no interface on subnet' in neutron_err:
                    #Attempted to delete a connection that does not exist.
                    #Ignore this conflict.
                    return
                logger.exception("Problem deleting interface router"
                             " from router %s to subnet %s."
                             % (router_id, subnet_id))
                raise

    def delete_router(self, neutron, router_name):
        router_id = self.get_router_id(self.neutron, router_name)
        if router_id:
            try:
                return neutron.delete_router(router_id)
            except:
                logger.error("Problem deleting router: %s" % router_id)
                raise

    def delete_subnet(self, neutron, subnet_name):
        subnet_id = self.get_subnet_id(neutron, subnet_name)
        if subnet_id:
            try:
                return neutron.delete_subnet(subnet_id)
            except:
                logger.error("Problem deleting subnet: %s" % subnet_id)
                raise

    def delete_network(self, neutron, network_name):
        network_id = self.get_network_id(neutron, network_name)
        if network_id:
            try:
                return neutron.delete_network(network_id)
            except:
                logger.error("Problem deleting network: %s" % network_id)
                raise

    def delete_port(self, port):
        return self.neutron.delete_port(port['id'])



"""
Atmosphere service meta.

DEPRECATION NOTE: This file should be removed. All relevant/useful work should be transferred out of this file.
"""
from abc import ABCMeta
from math import floor
import sys
import urlparse

from threepio import logger

from rtwo import settings

from rtwo.models.provider import AWSProvider, EucaProvider, OSProvider,\
    OSValhallaProvider
from rtwo.models.identity import AWSIdentity, EucaIdentity, OSIdentity
from rtwo.driver import AWSDriver, EucaDriver, OSDriver
from rtwo.linktest import active_instances

from rtwo.accounts.openstack import AccountDriver as OSAccountDriver


class BaseMeta(object):
    __metaclass__ = ABCMeta


class Meta(BaseMeta):

    provider = None

    metas = {}

    def __init__(self, driver, admin_driver=None):
        self._driver = driver._connection
        self.user = driver.identity.user
        self.provider = driver.provider
        self.provider_options = driver.provider.options
        self.identity = driver.identity
        self.driver = driver
        if not admin_driver:
            self.admin_driver = self.create_admin_driver({})
        else:
            self.admin_driver = admin_driver

    @classmethod
    def create_meta(cls, driver, admin_driver=None):
        meta = driver.provider.metaCls(driver, admin_driver)
        cls.metas[(driver.provider, driver.identity)] = meta
        return meta

    @classmethod
    def get_meta(cls, driver):
        id = (cls.provider, driver.identity)
        if cls.metas.get(id):
            return cls.metas[id]
        else:
            return cls.create_meta(driver)

    @classmethod
    def get_metas(cls):
        super_metas = {}
        map(super_metas.update, [AWSProvider.metaCls.metas,
                                 EucaProvider.metaCls.metas,
                                 OSProvider.metaCls.metas])
        return super_metas

    def test_links(self):
        return active_instances(self.driver.list_instances())

    def _split_creds(self, creds, default_key,
                     default_secret, default_tenant=None):
        key = creds.get('key', default_key)
        secret = creds.get('secret', default_secret)
        # Use the project or tenant name.
        tenant = creds.get('ex_project_name',
                           creds.get('ex_tenant_name', default_tenant))
        if tenant:
            return (key, secret, tenant)
        else:
            return (key, secret)

    def create_admin_driver(self, creds=None):
        raise NotImplementedError

    def all_instances(self):
        return self.provider.instanceCls.get_instances(
            self.admin_driver._connection.ex_list_all_instances(),
            self.provider)

    def reset(self):
        Meta.reset()
        self.metas = {}

    @classmethod  # order matters... /sigh
    def reset(cls):
        cls.metas = {}

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return reduce(lambda x, y: x+y, map(unicode, [self.__class__,
                                                      " ",
                                                      self.json()]))

    def __repr__(self):
        return str(self)

    def json(self):
        return {'driver': self.driver,
                'identity': self.identity,
                'provider': self.provider.name}


class AWSMeta(Meta):

    provider = AWSProvider

    def create_admin_driver(self, creds=None):
        if not hasattr(settings, 'AWS_KEY'):
            return self.driver
        logger.debug(self.provider)
        logger.debug(type(self.provider))
        identity = AWSIdentity(self.provider,
                               settings.AWS_KEY,
                               settings.AWS_SECRET)
        driver = AWSDriver(self.provider, identity)
        return driver

    def all_instances(self):
        return self.admin_driver.list_instances()


class EucaMeta(Meta):

    provider = EucaProvider

    def create_admin_driver(self, creds=None):
        key, secret = self._split_creds(creds,
                                        settings.EUCA_ADMIN_KEY,
                                        settings.EUCA_ADMIN_SECRET)
        identity = EucaIdentity(self.provider, key, secret)
        driver = EucaDriver(self.provider, identity)
        return driver

    def occupancy(self):
        return self.admin_driver.list_sizes()

    def all_instances(self):
        return self.admin_driver.list_instances()


class OSMeta(Meta):

    provider = OSProvider

    def create_admin_driver(self, creds=None):
        """
        DEPRECATED
        """
        admin_provider = OSProvider()
        provider_creds = self.provider_options
        key, secret, tenant =\
            self._split_creds(creds,
                              '',
                              '',
                              '')
        admin_identity = OSIdentity(admin_provider,
                                    key,
                                    secret,
                                    ex_tenant_name=tenant)
        admin_driver = OSDriver(admin_provider,
                                admin_identity,
                                **provider_creds)
        return admin_driver

    def total_remaining(self, max_, total, used, size):
        """
        Given max, total used and size calculate and return a
        2-tuple with the (max, remaining).
        """
        if size != 0:
            return (floor(max_), floor((total - used) / size))
        else:
            return (floor(max_), sys.maxint)

    def _cpu_stats(self, size, cpu_total, cpu_used, cpu_overcommit):
        if cpu_overcommit > 0:
            cpu_used = cpu_used - cpu_overcommit
        # CPUs go by many different, provider-specific names..
        if hasattr(size._size, 'cpu'):
            cpu_count = size._size.cpu
        elif hasattr(size._size, 'vcpus'):
            cpu_count = size._size.vcpus
        else:
            logger.warn("Could not find a CPU value for size %s" % size)
            cpu_count = -1
        if cpu_count > 0:
            max_by_cpu = float(cpu_total)/float(size.cpu)
        else:
            # I don't know about this?
            max_by_cpu = sys.maxint
        return self.total_remaining(max_by_cpu, cpu_total,
                                    cpu_used, cpu_count)

    def _ram_stats(self, size, ram_total, ram_used, ram_overcommit):
        if ram_overcommit > 0:
            ram_used = ram_used - ram_overcommit
        if size._size.ram > 0:
            max_by_ram = float(ram_total) / float(size._size.ram)
        else:
            max_by_ram = sys.maxint
        return self.total_remaining(max_by_ram, ram_total, ram_used, size.ram)

    def _disk_stats(self, size, disk_total, disk_used, disk_overcommit):
        if disk_overcommit > 0:
            disk_used = disk_used - disk_overcommit
        if size._size.disk > 0:
            max_by_disk = float(disk_total) / float(size._size.disk)
        else:
            max_by_disk = sys.maxint
        return self.total_remaining(max_by_disk, disk_total,
                                    disk_used, size._size.disk)

    def _calculate_overcommits(self, sizes, remove_totals):
        instances = self.admin_driver.list_all_instances()
        size_map = {size.id:size for size in sizes}
        for instance in instances:
            if instance.extra['status'] in ['suspended','shutoff']:
                #oc == OverCommited
                oc_size = size_map.get(instance.size.id)
                if not oc_size:
                    logger.warn("Size %s NOT found in list of sizes. Cannot"
                                " remove instance %s from calculation"
                                % (instance.size.id, instance.id))
                    continue
                remove_totals['cpu'] = remove_totals['cpu'] + oc_size.cpu
                remove_totals['ram'] = remove_totals['ram'] + oc_size.ram
                remove_totals['disk'] = remove_totals['disk'] + oc_size.disk
        return remove_totals


    def _instance_capacity_on_node(self, size, node):
        pass

    def _instance_capacity_vcpus(self):
        pass

    def _sum_active_compute_nodes(self):
        acs = self._active_compute_nodes()
        return {"total_vcpus": sum([ac["vcpus"] for ac in acs]),
                "total_memory_mb": sum([ac["memory_mb"] for ac in acs]),
                "total_local_gb": sum(ac["local_gb"] for ac in acs)}

    def _active_compute_nodes(self):
        active_nodes = set(
            [self._scrub_hostname(n["host"])\
             for n in self.admin_driver._connection.ex_os_services()\
             if n["status"] == "enabled"])
        nodes = {}
        for n in self.admin_driver._connection.ex_detail_hypervisor_nodes():
            hostname = self._scrub_hostname(n["hypervisor_hostname"])
            if hostname in active_nodes:
                nodes[hostname] = n
        return nodes

    def _get_node(self, nodes, instance):
        hostname = instance\
            .extra["object"]\
            .get("OS-EXT-SRV-ATTR:hypervisor_hostname")
        if hostname:
            hostname = self._scrub_hostname(hostname)
        return nodes.get(hostname)

    def _get_size(self, sizes, instance):
        size_id = instance.size.id
        size = [s for s in sizes if s.id == size_id]
        if size and len(size) == 1:
            return size[0]
        else:
            return None

    def _scrub_hostname(self, hostname):
        if not hostname:
            return None
        url = urlparse.urlparse(hostname)
        if not url.hostname:
            return url.path.split(".")[0]
        else:
            return url.hostname.split(".")[0]

    def _get_hashable_node(self, node):
        if not node:
            return None
        return self._scrub_hostname(node["hypervisor_hostname"])

    def _add_occupancy(self, occupancy, node, size, instance):
        node_key = self._get_hashable_node(node)
        occ = occupancy.get(node_key)
        if not occ:
            occupancy[node_key] = {
                "cpu": 0,
                "mem": 0,
                "disk": 0,
                "instances": []
            }
            occ = occupancy[node_key]
        occ["cpu"] += size.cpu
        occ["mem"] += size.ram
        occ["disk"] += size.ephemeral
        occ["instances"].append(instance)

# for host, stats in rahr.items():
#         if not host:
#             continue
#         node = nodes[host]
#         max_cpu = node["vcpus"]
#         max_mem = node["memory_mb"]
#         print("host[{0}] cpu ratio({1}/{2}vcpus): {3:03.5f} memory ratio ({4}/{5}MB): {6:03.5f}".format(
#             host,
#             stats["cpu"],
#             max_cpu,
#             stats["cpu"]/(1.0*max_cpu),
#             stats["mem"],
#             max_mem,
#             stats["mem"]/(1.0*max_mem)))


    def new_occupancy(self, overcommited=True):
        """
        Calculate occupancy using an admin account.

        Get size, instance and compute node data then
        calculate a better occupancy than _ex_hypervisor_statistics.
        Our statistics only look at instance states that use resources
        on the compute node.
        """
        sizes = self.admin_driver.list_sizes()
        nodes = self._active_compute_nodes()
        occupancy = {}
        for i in self.all_instances():
            node = self._get_node(nodes, i)
            size = self._get_size(sizes, i)
            self._add_occupancy(occupancy, node, size, i)
        return occupancy

    def occupancy(self, overcommited=False):
        """
        Add Occupancy data to NodeSize.extra
        """
        occ = self.admin_driver._connection\
                               .ex_hypervisor_statistics()
        remove_totals = {
                'cpu':0,
                'ram':0,
                'disk':0
            }
        sizes = self.admin_driver.list_sizes()
        if not overcommited:
            self._calculate_overcommits(sizes, remove_totals)

        for size in sizes:
            total_cpu, remaining_cpu = self._cpu_stats(size,
                                                      occ['vcpus'],
                                                      occ['vcpus_used'],
                                                      remove_totals['cpu'])
        
            total_ram, remaining_ram = self._ram_stats(size,
                                                       occ['memory_mb'],
                                                       occ['memory_mb_used'],
                                                       remove_totals['ram'])
            total_disk, remaining_disk = self._disk_stats(size,
                                                         occ['local_gb'],
                                                         occ['local_gb_used'],
                                                         remove_totals['disk'])
            remaining = min(remaining_cpu,
                            remaining_ram,
                            remaining_disk)
            if remaining == remaining_cpu:
                total = total_cpu
            elif remaining == remaining_ram:
                total = total_ram
            else:
                total = total_disk
            size.extra['occupancy'] = {'total': total,
                                       'remaining': remaining}
        return sizes

    def add_metadata_deployed(self, machine):
        """
        Add {"deployed": "True"} key and value to the machine's metadata.
        """
        machine_metadata = self.admin_driver._connection\
                                            .ex_get_image_metadata(machine)
        machine_metadata["deployed"] = "True"
        self.admin_driver._connection\
                         .ex_set_image_metadata(machine, machine_metadata)

    def remove_metadata_deployed(self, machine):
        """
        Remove the {"deployed": "True"} key and value from the machine's
        metadata, if it exists.
        """
        machine_metadata = self.admin_driver._connection\
                                            .ex_get_image_metadata(machine)
        if machine_metadata.get("deployed"):
            self.admin_driver._connection.ex_delete_image_metadata(machine,
                                                                   "deployed")

    def stop_all_instances(self, destroy=False):
        """
        Stop all instances and delete tenant networks for all users.

        To destroy instances instead of stopping them use the destroy
        keyword (destroy=True).
        """
        for instance in self.all_instances():
            if destroy:
                self.admin_driver.destroy_instance(instance)
                logger.debug('Destroyed instance %s' % instance)
            else:
                if instance.get_status() == 'active':
                    self.admin_driver.stop_instance(instance)
                    logger.debug('Stopped instance %s' % instance)
        os_driver = OSAccountDriver()
        if destroy:
            for username in os_driver.list_usergroup_names():
                tenant_name = username
                os_driver.network_manager.delete_tenant_network(username,
                                                                tenant_name)
        return True

    def destroy_all_instances(self):
        """
        Destroy all instances and delete tenant networks for all users.
        """
        for instance in self.all_instances():
            self.admin_driver.destroy_instance(instance)
            logger.debug('Destroyed instance %s' % instance)
        os_driver = OSAccountDriver()
        for username in os_driver.list_usergroup_names():
            tenant_name = username
            os_driver.network_manager.delete_tenant_network(username,
                                                            tenant_name)
        return True

    def all_instances(self, **kwargs):
        return self.provider.instanceCls.get_instances(
            self.admin_driver._connection.ex_list_all_instances(**kwargs),
            self.provider)

    def all_volumes(self):
        return self.provider.instanceCls.get_volumes(
            self.admin_driver._connection.ex_list_all_volumes())
